"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import itertools
import ntpath
import posixpath
import re
from pathlib import Path
from typing import Optional, Pattern, Tuple

from ._base import DEFAULT_MIN_LEN, AbstractSanitizer, BaseFile, BaseValidator
from ._common import (
    PathType,
    Platform,
    PlatformType,
    findall_to_str,
    is_pathlike_obj,
    preprocess,
    validate_pathtype,
)
from .error import ErrorReason, InvalidCharError, InvalidLengthError, ValidationError
from .handler import Handler


_DEFAULT_MAX_FILENAME_LEN = 255
_RE_INVALID_FILENAME = re.compile(f"[{re.escape(BaseFile._INVALID_FILENAME_CHARS):s}]", re.UNICODE)
_RE_INVALID_WIN_FILENAME = re.compile(
    f"[{re.escape(BaseFile._INVALID_WIN_FILENAME_CHARS):s}]", re.UNICODE
)


class FileNameSanitizer(AbstractSanitizer):
    def __init__(
        self,
        min_len: int = DEFAULT_MIN_LEN,
        max_len: int = _DEFAULT_MAX_FILENAME_LEN,
        platform: PlatformType = None,
        check_reserved: bool = True,
        null_value_handler: Optional[Handler] = None,
    ) -> None:
        super().__init__(
            min_len=min_len,
            max_len=max_len,
            check_reserved=check_reserved,
            null_value_handler=null_value_handler,
            platform_max_len=_DEFAULT_MAX_FILENAME_LEN,
            platform=platform,
        )

        self._sanitize_regexp = self._get_sanitize_regexp()
        self.__validator = FileNameValidator(
            min_len=self.min_len,
            max_len=self.max_len,
            check_reserved=check_reserved,
            platform=self.platform,
        )

    def sanitize(self, value: PathType, replacement_text: str = "") -> PathType:
        try:
            validate_pathtype(value, allow_whitespaces=False if self._is_windows() else True)
        except ValidationError as e:
            if e.reason == ErrorReason.NULL_NAME:
                return self._null_value_handler(e)
            raise

        sanitized_filename = self._sanitize_regexp.sub(replacement_text, str(value))
        sanitized_filename = sanitized_filename[: self.max_len]

        try:
            self.__validator.validate(sanitized_filename)
        except ValidationError as e:
            if e.reason == ErrorReason.RESERVED_NAME and e.reusable_name is False:
                sanitized_filename = re.sub(
                    re.escape(e.reserved_name), f"{e.reserved_name}_", sanitized_filename
                )
            elif e.reason == ErrorReason.INVALID_CHARACTER:
                if self.platform in [Platform.UNIVERSAL, Platform.WINDOWS]:
                    # Do not end a file or directory name with a space or a period
                    sanitized_filename = sanitized_filename.rstrip(" .")
            elif e.reason == ErrorReason.NULL_NAME:
                return self._null_value_handler(e)
            else:
                raise

        if is_pathlike_obj(value):
            return Path(sanitized_filename)

        return sanitized_filename

    def _get_sanitize_regexp(self) -> Pattern:
        if self.platform in [Platform.UNIVERSAL, Platform.WINDOWS]:
            return _RE_INVALID_WIN_FILENAME

        return _RE_INVALID_FILENAME


class FileNameValidator(BaseValidator):
    _WINDOWS_RESERVED_FILE_NAMES = ("CON", "PRN", "AUX", "CLOCK$", "NUL") + tuple(
        f"{name:s}{num:d}" for name, num in itertools.product(("COM", "LPT"), range(1, 10))
    )
    _MACOS_RESERVED_FILE_NAMES = (":",)

    @property
    def reserved_keywords(self) -> Tuple[str, ...]:
        common_keywords = super().reserved_keywords

        if self._is_universal():
            return (
                common_keywords
                + self._WINDOWS_RESERVED_FILE_NAMES
                + self._MACOS_RESERVED_FILE_NAMES
            )

        if self._is_windows():
            return common_keywords + self._WINDOWS_RESERVED_FILE_NAMES

        if self._is_posix() or self._is_macos():
            return common_keywords + self._MACOS_RESERVED_FILE_NAMES

        return common_keywords

    def __init__(
        self,
        min_len: int = DEFAULT_MIN_LEN,
        max_len: int = _DEFAULT_MAX_FILENAME_LEN,
        platform: PlatformType = None,
        check_reserved: bool = True,
    ) -> None:
        super().__init__(
            min_len=min_len,
            max_len=max_len,
            check_reserved=check_reserved,
            platform_max_len=_DEFAULT_MAX_FILENAME_LEN,
            platform=platform,
        )

    def validate(self, value: PathType) -> None:
        validate_pathtype(
            value,
            allow_whitespaces=False
            if self.platform in [Platform.UNIVERSAL, Platform.WINDOWS]
            else True,
        )

        unicode_filename = preprocess(value)
        value_len = len(unicode_filename)

        self.validate_abspath(unicode_filename)

        if value_len > self.max_len:
            raise InvalidLengthError(
                f"filename is too long: expected<={self.max_len:d}, actual={value_len:d}"
            )
        if value_len < self.min_len:
            raise InvalidLengthError(
                f"filename is too short: expected>={self.min_len:d}, actual={value_len:d}"
            )

        self._validate_reserved_keywords(unicode_filename)

        if self._is_universal() or self._is_windows():
            self.__validate_win_filename(unicode_filename)
        else:
            self.__validate_unix_filename(unicode_filename)

    def validate_abspath(self, value: str) -> None:
        err = ValidationError(
            description=f"found an absolute path ({value}), expected a filename",
            platform=self.platform,
            reason=ErrorReason.FOUND_ABS_PATH,
        )

        if self._is_universal() or self._is_windows():
            if ntpath.isabs(value):
                raise err

        if posixpath.isabs(value):
            raise err

    def __validate_unix_filename(self, unicode_filename: str) -> None:
        match = _RE_INVALID_FILENAME.findall(unicode_filename)
        if match:
            raise InvalidCharError(
                self._ERROR_MSG_TEMPLATE.format(
                    invalid=findall_to_str(match), value=repr(unicode_filename)
                )
            )

    def __validate_win_filename(self, unicode_filename: str) -> None:
        match = _RE_INVALID_WIN_FILENAME.findall(unicode_filename)
        if match:
            raise InvalidCharError(
                self._ERROR_MSG_TEMPLATE.format(
                    invalid=findall_to_str(match), value=repr(unicode_filename)
                ),
                platform=Platform.WINDOWS,
            )

        if unicode_filename in (".", ".."):
            return

        if unicode_filename[-1] in (" ", "."):
            raise InvalidCharError(
                self._ERROR_MSG_TEMPLATE.format(
                    invalid=re.escape(unicode_filename[-1]), value=repr(unicode_filename)
                ),
                platform=Platform.WINDOWS,
                description="Do not end a file or directory name with a space or a period",
            )


def validate_filename(
    filename: PathType,
    platform: Optional[str] = None,
    min_len: int = DEFAULT_MIN_LEN,
    max_len: int = _DEFAULT_MAX_FILENAME_LEN,
    check_reserved: bool = True,
) -> None:
    """Verifying whether the ``filename`` is a valid file name or not.

    Args:
        filename:
            Filename to validate.
        platform:
            Target platform name of the filename.

            .. include:: platform.txt
        min_len:
            Minimum length of the ``filename``. The value must be greater or equal to one.
            Defaults to ``1``.
        max_len:
            Maximum length of the ``filename``. The value must be lower than:

                - ``Linux``: 4096
                - ``macOS``: 1024
                - ``Windows``: 260
                - ``universal``: 260

            Defaults to ``255``.
        check_reserved:
            If |True|, check reserved names of the ``platform``.

    Raises:
        ValidationError (ErrorReason.INVALID_LENGTH):
            If the ``filename`` is longer than ``max_len`` characters.
        ValidationError (ErrorReason.INVALID_CHARACTER):
            If the ``filename`` includes invalid character(s) for a filename:
            |invalid_filename_chars|.
            The following characters are also invalid for Windows platform:
            |invalid_win_filename_chars|.
        ValidationError (ErrorReason.RESERVED_NAME):
            If the ``filename`` equals reserved name by OS.
            Windows reserved name is as follows:
            ``"CON"``, ``"PRN"``, ``"AUX"``, ``"NUL"``, ``"COM[1-9]"``, ``"LPT[1-9]"``.

    Example:
        :ref:`example-validate-filename`

    See Also:
        `Naming Files, Paths, and Namespaces - Win32 apps | Microsoft Docs
        <https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file>`__
    """

    FileNameValidator(
        platform=platform, min_len=min_len, max_len=max_len, check_reserved=check_reserved
    ).validate(filename)


def is_valid_filename(
    filename: PathType,
    platform: Optional[str] = None,
    min_len: int = DEFAULT_MIN_LEN,
    max_len: Optional[int] = None,
    check_reserved: bool = True,
) -> bool:
    """Check whether the ``filename`` is a valid name or not.

    Args:
        filename:
            A filename to be checked.

    Example:
        :ref:`example-is-valid-filename`

    See Also:
        :py:func:`.validate_filename()`
    """

    return FileNameValidator(
        platform=platform,
        min_len=min_len,
        max_len=-1 if max_len is None else max_len,
        check_reserved=check_reserved,
    ).is_valid(filename)


def sanitize_filename(
    filename: PathType,
    replacement_text: str = "",
    platform: Optional[str] = None,
    max_len: Optional[int] = _DEFAULT_MAX_FILENAME_LEN,
    check_reserved: bool = True,
    null_value_handler: Optional[Handler] = None,
) -> PathType:
    """Make a valid filename from a string.

    To make a valid filename the function does:

        - Replace invalid characters as file names included in the ``filename``
          with the ``replacement_text``. Invalid characters are:

            - unprintable characters
            - |invalid_filename_chars|
            - for Windows (or universal) only: |invalid_win_filename_chars|

        - Append underscore (``"_"``) at the tail of the name if sanitized name
          is one of the reserved names by operating systems
          (only when ``check_reserved`` is |True|).

    Args:
        filename: Filename to sanitize.
        replacement_text:
            Replacement text for invalid characters. Defaults to ``""``.
        platform:
            Target platform name of the filename.

            .. include:: platform.txt
        max_len:
            Maximum length of the ``filename`` length. Truncate the name length if
            the ``filename`` length exceeds this value.
            Defaults to ``255``.
        check_reserved:
            If |True|, sanitize reserved names of the ``platform``.
        null_value_handler:
            Function called when a value after sanitization is an empty string.
            Defaults to ``pathvalidate.handler.return_null_string()`` that just return ``""``.

    Returns:
        Same type as the ``filename`` (str or PathLike object):
            Sanitized filename.

    Raises:
        ValueError:
            If the ``filename`` is an invalid filename.

    Example:
        :ref:`example-sanitize-filename`
    """

    return FileNameSanitizer(
        platform=platform,
        max_len=-1 if max_len is None else max_len,
        check_reserved=check_reserved,
        null_value_handler=null_value_handler,
    ).sanitize(filename, replacement_text)



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import abc

from ._interface import TypeCheckerInterface


class CheckerFactory:
    __slots__ = ("__min_strict_level", "__max_strict_level", "__checker_mapping")

    @property
    def min_strict_level(self):
        return self.__min_strict_level

    @property
    def max_strict_level(self):
        return self.__max_strict_level

    def __init__(self, checker_mapping):
        self.__checker_mapping = checker_mapping

        self.__min_strict_level = min(checker_mapping)
        self.__max_strict_level = max(checker_mapping)
        self.__checker_mapping[None] = self.__max_strict_level

    def get_checker_class(self, strict_level=None):
        checker_class = self.__checker_mapping.get(strict_level)
        if checker_class:
            return checker_class
        if strict_level < self.min_strict_level:
            return self.__checker_mapping[self.min_strict_level]
        if strict_level > self.max_strict_level:
            return self.__checker_mapping[self.max_strict_level]

        raise ValueError(f"unexpected strict level: {strict_level}")


class TypeCheckerBase(TypeCheckerInterface):
    __slots__ = ("_value",)

    def __init__(self, value):
        self._value = value

    @abc.abstractmethod
    def is_instance(self):
        pass

    def is_type(self) -> bool:
        return self.is_instance() and not self.is_exclude_instance()

    def validate(self) -> None:
        """
        :raises TypeError:
            If the value is not matched the type to be expected.
        """

        if self.is_type():
            return

        raise TypeError(f"invalid value type: actual={type(self._value)}")

    def is_exclude_instance(self):
        return False

    def is_valid_after_convert(self, converted_value):
        return True


class TypeCheckerDelegator(TypeCheckerInterface):
    __slots__ = ("__checker",)

    def __init__(self, value, checker_factory, strict_level):
        self.__checker = checker_factory.get_checker_class(strict_level)(value)

    def is_type(self) -> bool:
        return self.__checker.is_type()

    def is_valid_after_convert(self, value):
        return self.__checker.is_valid_after_convert(value)

    def is_instance(self):
        return self.__checker.is_instance()

    def is_exclude_instance(self):
        return self.__checker.is_exclude_instance()

    def validate(self) -> None:
        self.__checker.validate()



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import os
from itertools import chain
from typing import List

import typepy

from ._logger import logger
from ._table_format import FormatAttr, TableFormat
from .error import WriterNotFoundError
from .writer._table_writer import AbstractTableWriter


class TableWriterFactory:
    """
    A factor class of table writer classes.
    """

    @classmethod
    def create_from_file_extension(cls, file_extension: str, **kwargs) -> AbstractTableWriter:
        """
        Create a table writer class instance from a file extension.
        Supported file extensions are as follows:

            ==================  ===================================
            Extension           Writer Class
            ==================  ===================================
            ``".adoc"``         :py:class:`~.AsciiDocTableWriter`
            ``".asciidoc"``     :py:class:`~.AsciiDocTableWriter`
            ``".asc"``          :py:class:`~.AsciiDocTableWriter`
            ``".css"``          :py:class:`~.CssTableWriter`
            ``".csv"``          :py:class:`~.CsvTableWriter`
            ``".htm"``          :py:class:`~.HtmlTableWriter`
            ``".html"``         :py:class:`~.HtmlTableWriter`
            ``".js"``           :py:class:`~.JavaScriptTableWriter`
            ``".json"``         :py:class:`~.JsonTableWriter`
            ``".jsonl"``        :py:class:`~.JsonLinesTableWriter`
            ``".ltsv"``         :py:class:`~.LtsvTableWriter`
            ``".ldjson"``       :py:class:`~.JsonLinesTableWriter`
            ``".md"``           :py:class:`~.MarkdownTableWriter`
            ``".ndjson"``       :py:class:`~.JsonLinesTableWriter`
            ``".py"``           :py:class:`~.PythonCodeTableWriter`
            ``".rst"``          :py:class:`~.RstGridTableWriter`
            ``".tsv"``          :py:class:`~.TsvTableWriter`
            ``".xls"``          :py:class:`~.ExcelXlsTableWriter`
            ``".xlsx"``         :py:class:`~.ExcelXlsxTableWriter`
            ``".sqlite"``       :py:class:`~.SqliteTableWriter`
            ``".sqlite3"``      :py:class:`~.SqliteTableWriter`
            ``".tsv"``          :py:class:`~.TsvTableWriter`
            ``".toml"``         :py:class:`~.TomlTableWriter`
            ``".yml"``          :py:class:`~.YamlTableWriter`
            ==================  ===================================

        :param str file_extension:
            File extension string (case insensitive).
        :param kwargs:
            Keyword arguments that passing to writer class constructor.
        :return:
            Writer instance that coincides with the ``file_extension``.
        :rtype:
            :py:class:`~pytablewriter.writer._table_writer.TableWriterInterface`
        :raises pytablewriter.WriterNotFoundError:
            |WriterNotFoundError_desc| the file extension.
        """

        ext = os.path.splitext(file_extension)[1]
        if typepy.is_null_string(ext):
            file_extension = file_extension
        else:
            file_extension = ext

        file_extension = file_extension.lstrip(".").lower()

        for table_format in TableFormat:
            if file_extension not in table_format.file_extensions:
                continue

            if table_format.format_attribute & FormatAttr.SECONDARY_EXT:
                continue

            logger.debug(f"create a {table_format.writer_class.__name__} instance")

            return table_format.writer_class(**kwargs)

        raise WriterNotFoundError(
            "\n".join(
                [
                    f"{file_extension:s} (unknown file extension).",
                    "",
                    "acceptable file extensions are: {}.".format(", ".join(cls.get_extensions())),
                ]
            )
        )

    @classmethod
    def create_from_format_name(cls, format_name: str, **kwargs) -> AbstractTableWriter:
        """
        Create a table writer class instance from a format name.
        Supported file format names are as follows:

            =============================================  ===================================
            Format name                                    Writer Class
            =============================================  ===================================
            ``"adoc"``                                     :py:class:`~.AsciiDocTableWriter`
            ``"asciidoc"``                                 :py:class:`~.AsciiDocTableWriter`
            ``"css"``                                      :py:class:`~.CssTableWriter`
            ``"csv"``                                      :py:class:`~.CsvTableWriter`
            ``"elasticsearch"``                            :py:class:`~.ElasticsearchWriter`
            ``"excel"``                                    :py:class:`~.ExcelXlsxTableWriter`
            ``"html"``/``"htm"``                           :py:class:`~.HtmlTableWriter`
            ``"javascript"``/``"js"``                      :py:class:`~.JavaScriptTableWriter`
            ``"json"``                                     :py:class:`~.JsonTableWriter`
            ``"json_lines"``                               :py:class:`~.JsonLinesTableWriter`
            ``"latex_matrix"``                             :py:class:`~.LatexMatrixWriter`
            ``"latex_table"``                              :py:class:`~.LatexTableWriter`
            ``"ldjson"``                                   :py:class:`~.JsonLinesTableWriter`
            ``"ltsv"``                                     :py:class:`~.LtsvTableWriter`
            ``"markdown"``/``"md"``                        :py:class:`~.MarkdownTableWriter`
            ``"mediawiki"``                                :py:class:`~.MediaWikiTableWriter`
            ``"null"``                                     :py:class:`~.NullTableWriter`
            ``"pandas"``                                   :py:class:`~.PandasDataFrameWriter`
            ``"py"``/``"python"``                          :py:class:`~.PythonCodeTableWriter`
            ``"rst"``/``"rst_grid"``/``"rst_grid_table"``  :py:class:`~.RstGridTableWriter`
            ``"rst_simple"``/``"rst_simple_table"``        :py:class:`~.RstSimpleTableWriter`
            ``"rst_csv"``/``"rst_csv_table"``              :py:class:`~.RstCsvTableWriter`
            ``"sqlite"``                                   :py:class:`~.SqliteTableWriter`
            ``"ssv"``                                      :py:class:`~.SpaceAlignedTableWriter`
            ``"tsv"``                                      :py:class:`~.TsvTableWriter`
            ``"toml"``                                     :py:class:`~.TomlTableWriter`
            ``"unicode"``                                  :py:class:`~.UnicodeTableWriter`
            ``"yaml"``                                     :py:class:`~.YamlTableWriter`
            =============================================  ===================================

        :param str format_name: Format name string (case insensitive).
        :param kwargs:
            Keyword arguments that passing to writer class constructor.
        :return: Writer instance that coincides with the ``format_name``:
        :rtype:
            :py:class:`~pytablewriter.writer._table_writer.TableWriterInterface`
        :raises pytablewriter.WriterNotFoundError:
            |WriterNotFoundError_desc| for the format.
        """

        format_name = format_name.casefold()

        for table_format in TableFormat:
            if format_name in table_format.names and not (
                table_format.format_attribute & FormatAttr.SECONDARY_NAME
            ):
                logger.debug(f"create a {table_format.writer_class.__name__} instance")

                return table_format.writer_class(**kwargs)

        raise WriterNotFoundError(
            "\n".join(
                [
                    f"{format_name} (unknown format name).",
                    "acceptable format names are: {}.".format(", ".join(cls.get_format_names())),
                ]
            )
        )

    @classmethod
    def get_format_names(cls) -> List[str]:
        """
        :return: Available format names.
        :rtype: list

        :Example:
            .. code:: python

                >>> import pytablewriter as ptw
                >>> for name in ptw.TableWriterFactory.get_format_names():
                ...     print(name)
                ...
                adoc
                asciidoc
                bold_unicode
                borderless
                css
                csv
                elasticsearch
                excel
                htm
                html
                javascript
                js
                json
                json_lines
                jsonl
                latex_matrix
                latex_table
                ldjson
                ltsv
                markdown
                md
                mediawiki
                ndjson
                null
                numpy
                pandas
                pandas_pickle
                py
                python
                rst
                rst_csv
                rst_csv_table
                rst_grid
                rst_grid_table
                rst_simple
                rst_simple_table
                space_aligned
                sqlite
                ssv
                toml
                tsv
                unicode
                yaml

        """

        return sorted(list(set(chain(*(table_format.names for table_format in TableFormat)))))

    @classmethod
    def get_extensions(cls) -> List[str]:
        """
        :return: Available file extensions.
        :rtype: list

        :Example:
            .. code:: python

                >>> import pytablewriter as ptw
                >>> for name in ptw.TableWriterFactory.get_extensions():
                ...     print(name)
                ...
                adoc
                asc
                asciidoc
                css
                csv
                htm
                html
                js
                json
                jsonl
                ldjson
                ltsv
                md
                ndjson
                py
                rst
                sqlite
                sqlite3
                tex
                toml
                tsv
                xls
                xlsx
                yml
        """

        file_extension_set = set()
        for table_format in TableFormat:
            for file_extension in table_format.file_extensions:
                file_extension_set.add(file_extension)

        return sorted(list(file_extension_set))



from typing import List

import typepy

from ._python import PythonCodeTableWriter


class NumpyTableWriter(PythonCodeTableWriter):
    """
    A table writer class for ``NumPy`` source code format.

        :Example:
            :ref:`example-numpy-table-writer`

    .. py:method:: write_table

        |write_table| with ``NumPy`` format.
        The tabular data are written as a variable definition of
        ``numpy.array``.

        :raises pytablewriter.EmptyTableNameError:
            If the |table_name| is empty.
        :Example:
            :ref:`example-numpy-table-writer`

        .. note::
            Specific values in the tabular data are converted when writing:

            - |None|: written as ``None``
            - |inf|: written as ``numpy.inf``
            - |nan|: written as ``numpy.nan``
            - |datetime| instances determined by |is_datetime_instance_formatting| attribute:
                - |True|: written as `dateutil.parser <https://dateutil.readthedocs.io/en/stable/parser.html>`__
                - |False|: written as |str|

            .. seealso::
                :ref:`example-type-hint-python`
    """

    FORMAT_NAME = "numpy"

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.import_numpy_as = "np"
        self._dp_extractor.type_value_map[typepy.Typecode.INFINITY] = "{:s}.inf".format(
            self.import_numpy_as
        )
        self._dp_extractor.type_value_map[typepy.Typecode.NAN] = "{:s}.nan".format(
            self.import_numpy_as
        )

    def _get_opening_row_items(self) -> List[str]:
        array_def = f"{self.import_numpy_as:s}.array(["

        if typepy.is_not_null_string(self.table_name):
            return [f"{self.variable_name} = {array_def}"]

        return [array_def]

    def _get_closing_row_items(self) -> List[str]:
        return ["])"]



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from ._checker import CheckerFactory, TypeCheckerBase, TypeCheckerDelegator
from ._common import isstring


class ListTypeCheckerStrictLevel0(TypeCheckerBase):
    def is_instance(self):
        return isinstance(self._value, list)

    def is_valid_after_convert(self, converted_value):
        return isinstance(converted_value, list) and converted_value

    def is_exclude_instance(self):
        return isstring(self._value)


class ListTypeCheckerStrictLevel1(ListTypeCheckerStrictLevel0):
    def is_exclude_instance(self):
        return super().is_exclude_instance() or not isinstance(self._value, list)


_factory = CheckerFactory(
    checker_mapping={0: ListTypeCheckerStrictLevel0, 1: ListTypeCheckerStrictLevel1}
)


class ListTypeChecker(TypeCheckerDelegator):
    def __init__(self, value, strict_level):
        super().__init__(value=value, checker_factory=_factory, strict_level=strict_level)



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""


import copy
from typing import Dict, Generator

import dataproperty
from dataproperty import ColumnDataProperty
from typepy import Typecode

from ..error import EmptyValueError
from ._msgfy import to_error_message
from ._table_writer import AbstractTableWriter


def _get_es_datatype(column_dp: ColumnDataProperty) -> Dict[str, str]:
    if column_dp.typecode in (
        Typecode.NONE,
        Typecode.NULL_STRING,
        Typecode.INFINITY,
        Typecode.NAN,
    ):
        return {"type": "keyword"}

    if column_dp.typecode == Typecode.STRING:
        return {"type": "text"}

    if column_dp.typecode == Typecode.DATETIME:
        return {"type": "date", "format": "date_optional_time"}

    if column_dp.typecode == Typecode.REAL_NUMBER:
        return {"type": "double"}

    if column_dp.typecode == Typecode.BOOL:
        return {"type": "boolean"}

    if column_dp.typecode == Typecode.IP_ADDRESS:
        return {"type": "ip"}

    if column_dp.typecode == Typecode.INTEGER:
        assert column_dp.bit_length is not None

        if column_dp.bit_length <= 8:
            return {"type": "byte"}
        elif column_dp.bit_length <= 16:
            return {"type": "short"}
        elif column_dp.bit_length <= 32:
            return {"type": "integer"}
        elif column_dp.bit_length <= 64:
            return {"type": "long"}

        raise ValueError(
            f"too large integer bits: expected<=64bits, actual={column_dp.bit_length:d}bits"
        )

    raise ValueError(f"unknown typecode: {column_dp.typecode}")


class ElasticsearchWriter(AbstractTableWriter):
    """
    A table writer class for Elasticsearch.

    :Dependency Packages:
        - `elasticsearch-py <https://github.com/elastic/elasticsearch-py>`__

    .. py:attribute:: index_name
        :type: str

        Alias attribute for |table_name|.

    .. py:attribute:: document_type
        :type: str
        :value: "table"

        Specify document type for indices.

    .. py:method:: write_table()

        Create an index and put documents for each row to Elasticsearch.

        You need to pass an
        `elasticsearch.Elasticsearch <https://elasticsearch-py.rtfd.io/en/master/api.html#elasticsearch>`__
        instance to |stream| before calling this method.
        |table_name|/:py:attr:`~pytablewriter.ElasticsearchWriter.index_name`
        used as the creating index name,
        invalid characters in the name are replaced with underscore (``'_'``).
        Document data types for documents are automatically detected
        from the data.

        :raises ValueError:
            If the |stream| has not elasticsearch.Elasticsearch instance.
        :Example:
            :ref:`example-elasticsearch-table-writer`
    """

    FORMAT_NAME = "elasticsearch"

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    @property
    def support_split_write(self) -> bool:
        return True

    @property
    def table_name(self) -> str:
        return super().table_name

    @table_name.setter
    def table_name(self, value) -> None:
        from pathvalidate import ErrorReason, ValidationError

        from ..sanitizer import ElasticsearchIndexNameSanitizer

        try:
            self._table_name = ElasticsearchIndexNameSanitizer(value).sanitize(replacement_text="_")
        except ValidationError as e:
            if e.reason is ErrorReason.NULL_NAME:
                self._table_name = ""
            else:
                raise

    @property
    def index_name(self) -> str:
        return self.table_name

    @index_name.setter
    def index_name(self, value: str) -> None:
        self.table_name = value

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.stream = None
        self.is_padding = False
        self.is_formatting_float = False
        self._is_require_table_name = True
        self._quoting_flags = copy.deepcopy(dataproperty.NOT_QUOTING_FLAGS)
        self._dp_extractor.type_value_map = copy.deepcopy(dataproperty.DefaultValue.TYPE_VALUE_MAP)

        self.document_type = "table"

    def write_null_line(self) -> None:
        pass

    def _get_mappings(self) -> Dict[str, Dict]:
        properties = {}

        for header, column_dp in zip(self.headers, self._column_dp_list):
            properties[header] = _get_es_datatype(column_dp)

        return {"mappings": {self.document_type: {"properties": properties}}}

    def _get_body(self) -> Generator:
        str_datatype = (Typecode.DATETIME, Typecode.IP_ADDRESS, Typecode.INFINITY, Typecode.NAN)

        for value_dp_list in self._table_value_dp_matrix:
            values = [
                value_dp.data if value_dp.typecode not in str_datatype else value_dp.to_str()
                for value_dp in value_dp_list
            ]

            yield dict(zip(self.headers, values))

    def _write_table(self, **kwargs) -> None:
        import elasticsearch as es

        if not isinstance(self.stream, es.Elasticsearch):
            raise ValueError("stream must be an elasticsearch.Elasticsearch instance")

        try:
            self._verify_value_matrix()
        except EmptyValueError:
            self._logger.logger.debug("no tabular data found")
            return

        self._preprocess()

        mappings = self._get_mappings()

        try:
            result = self.stream.indices.create(index=self.index_name, body=mappings)
            self._logger.logger.debug(result)
        except es.TransportError as e:
            if e.error == "index_already_exists_exception":
                # ignore already existing index
                self._logger.logger.debug(to_error_message(e))
            else:
                raise

        for body in self._get_body():
            try:
                self.stream.index(index=self.index_name, body=body, doc_type=self.document_type)
            except es.exceptions.RequestError as e:
                self._logger.logger.error(f"{to_error_message(e)}, body={body}")

    def _write_value_row_separator(self) -> None:
        pass



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import copy
import re
from typing import List, Optional, Sequence

from ._func import to_codec_name


def b(s: str) -> bytes:
    return s.encode("latin-1")


class MultiByteStrDecoder:
    """
    Reference:
        https://docs.python.org/3/library/codecs.html
    """

    __CODECS = [
        "utf_7",
        "utf_8",
        "utf_8_sig",
        "utf_16",
        "utf_16_be",
        "utf_16_le",
        "utf_32",
        "utf_32_be",
        "utf_32_le",
        "big5",
        "big5hkscs",
        "cp037",
        "cp424",
        "cp437",
        "cp500",
        "cp720",
        "cp737",
        "cp775",
        "cp850",
        "cp852",
        "cp855",
        "cp856",
        "cp857",
        "cp858",
        "cp860",
        "cp861",
        "cp862",
        "cp863",
        "cp864",
        "cp865",
        "cp866",
        "cp869",
        "cp874",
        "cp875",
        "cp932",
        "cp949",
        "cp950",
        "cp1006",
        "cp1026",
        "cp1140",
        "cp1250",
        "cp1251",
        "cp1252",
        "cp1253",
        "cp1254",
        "cp1255",
        "cp1256",
        "cp1257",
        "cp1258",
        "euc_jp",
        "euc_jis_2004",
        "euc_jisx0213",
        "euc_kr",
        "gb2312",
        "gbk",
        "gb18030",
        "hz",
        "iso2022_jp",
        "iso2022_jp_1",
        "iso2022_jp_2",
        "iso2022_jp_2004",
        "iso2022_jp_3",
        "iso2022_jp_ext",
        "iso2022_kr",
        "latin_1",
        "iso8859_2",
        "iso8859_3",
        "iso8859_4",
        "iso8859_5",
        "iso8859_6",
        "iso8859_7",
        "iso8859_8",
        "iso8859_9",
        "iso8859_10",
        "iso8859_11",
        "iso8859_13",
        "iso8859_14",
        "iso8859_15",
        "iso8859_16",
        "johab",
        "koi8_r",
        "koi8_u",
        "mac_cyrillic",
        "mac_greek",
        "mac_iceland",
        "mac_latin2",
        "mac_roman",
        "mac_turkish",
        "ptcp154",
        "shift_jis",
        "shift_jis_2004",
        "shift_jisx0213",
        "base64_codec",
        "bz2_codec",
        "hex_codec",
        "idna",
        "mbcs",
        "palmos",
        "punycode",
        "quopri_codec",
        "raw_unicode_escape",
        "rot_13",
        "string_escape",
        "unicode_escape",
        "unicode_internal",
        "uu_codec",
        "zlib_codec",
    ]
    __RE_UTF7 = re.compile(b("[+].*?[-]"))

    @property
    def unicode_str(self) -> str:
        return self.__unicode_str

    @property
    def codec(self) -> Optional[str]:
        return self.__codec

    def __init__(self, value, codec_candidates: Sequence[str] = None) -> None:
        self.__encoded_str = value
        self.__codec: Optional[str] = None
        if codec_candidates is None:
            self.__codec_candidate_list: List[str] = []
        else:
            self.__codec_candidate_list = list(codec_candidates)

        self.__validate_str()

        self.__unicode_str = self.__to_unicode()

    def __repr__(self) -> str:
        return f"codec={self.codec:s}, unicode={self.unicode_str:s}"

    def __validate_str(self) -> None:
        if isinstance(self.__encoded_str, (str, bytes)):
            return

        raise ValueError(f"value must be a string: actual={type(self.__encoded_str)}")

    def __is_buffer(self) -> bool:
        return isinstance(self.__encoded_str, memoryview)

    def __is_multibyte_utf7(self, encoded_str) -> bool:
        if self.__codec != "utf_7":
            return False

        utf7_symbol_count = encoded_str.count(b("+"))
        if utf7_symbol_count <= 0:
            return False

        if utf7_symbol_count != encoded_str.count(b("-")):
            return False

        return utf7_symbol_count == len(self.__RE_UTF7.findall(encoded_str))

    def __get_encoded_str(self) -> str:
        if self.__is_buffer():
            return str(self.__encoded_str)

        return self.__encoded_str

    @staticmethod
    def __detect_encoding_helper(encoded_str) -> Optional[str]:
        import chardet

        try:
            detect = chardet.detect(encoded_str)
        except TypeError:
            detect = {}

        detect_encoding = detect.get("encoding")
        confidence = detect.get("confidence")

        if detect_encoding not in ["ascii", "utf-8"] and confidence and confidence > 0.7:
            # utf7 tend to be misrecognized as ascii
            return detect_encoding

        return None

    def __get_codec_candidate_list(self, encoded_str) -> List[str]:
        codec_candidate_list = copy.deepcopy(self.__CODECS)
        detect_encoding = self.__detect_encoding_helper(encoded_str)

        if detect_encoding:
            try:
                codec_candidate_list.remove(detect_encoding)
            except ValueError:
                pass

            codec_candidate_list.insert(0, detect_encoding)

        for codec_candidate in self.__codec_candidate_list:
            try:
                codec_candidate_list.remove(codec_candidate)
            except ValueError:
                pass

        return self.__codec_candidate_list + codec_candidate_list

    def __to_unicode(self):
        encoded_str = self.__get_encoded_str()

        if encoded_str == b"":
            self.__codec = "unicode"
            return ""

        for codec in self.__get_codec_candidate_list(encoded_str):
            if not codec:
                continue

            try:
                self.__codec = to_codec_name(codec)
                decoded_str = encoded_str.decode(codec)
                break
            except UnicodeDecodeError:
                self.__codec = None
                continue
            except UnicodeEncodeError:
                # already a unicode string (python 2)
                self.__codec = "unicode"
                return encoded_str
            except AttributeError:
                if isinstance(encoded_str, str):
                    # already a unicode string (python 3)
                    self.__codec = "unicode"

                    if not encoded_str:
                        return encoded_str

                    try:
                        # python 2 compatibility
                        encoded_str.encode("ascii")
                        self.__codec = "ascii"
                        return encoded_str
                    except UnicodeEncodeError:
                        pass

                    return encoded_str

                self.__codec = None

                try:
                    return f"{encoded_str}"
                except UnicodeDecodeError:
                    # some of the objects that cannot convertible to a string
                    # may reach this line
                    raise TypeError("argument must be a string")
        else:
            self.__codec = None

            try:
                message = f"unknown codec: encoded_str={encoded_str}"
            except UnicodeDecodeError:
                message = f"unknown codec: value-type={type(encoded_str)}"

            raise UnicodeDecodeError(message)

        if self.codec == "utf_7":
            return self.__process_utf7(encoded_str, decoded_str)

        return decoded_str

    def __process_utf7(self, encoded_str, decoded_str) -> str:
        if not encoded_str:
            self.__codec = "unicode"

            return encoded_str

        if self.__is_multibyte_utf7(encoded_str):
            try:
                decoded_str.encode("ascii")

                self.__codec = "ascii"

                return encoded_str.decode("ascii")
            except UnicodeEncodeError:
                return decoded_str

        self.__codec = "ascii"

        return encoded_str.decode("ascii")



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from datetime import date, datetime

from .._common import strip_ansi_escape
from .._const import DefaultValue, ParamKey
from ..error import TypeConversionError
from ._interface import AbstractValueConverter


class DateTimeConverter(AbstractValueConverter):

    __DAYS_TO_SECONDS_COEF = 60 ** 2 * 24
    __MICROSECONDS_TO_SECONDS_COEF = 1000 ** 2
    __COMMON_DST_TIMEZONE_TABLE = {
        -36000: "America/Adak",  # -1000
        -32400: "US/Alaska",  # -0900
        -28800: "US/Pacific",  # -0800
        -25200: "US/Mountain",  # -0700
        -21600: "US/Central",  # -0600
        -18000: "US/Eastern",  # -0500
        -14400: "Canada/Atlantic",  # -0400
        -12600: "America/St_Johns",  # -0330
        -10800: "America/Miquelon",  # -0300
        7200: "Africa/Tripoli",  # 0200
    }

    def __init__(self, value, params):
        super().__init__(value, params)

        self.__datetime = None
        self.__timezone = self._params.get(ParamKey.TIMEZONE)

    def force_convert(self):
        self.__datetime = self.__from_datetime()
        if self.__datetime:
            return self.__datetime

        self.__datetime = self.__from_timestamp()
        if self.__datetime:
            return self.__datetime

        return self.__from_datetime_string()

    def __from_datetime(self):
        if not isinstance(self._value, (date, datetime)):
            return None

        if isinstance(self._value, datetime):
            self.__datetime = self._value
        elif isinstance(self._value, date):
            self.__datetime = datetime(
                year=self._value.year, month=self._value.month, day=self._value.day
            )

        if self.__timezone:
            self.__datetime = self.__timezone.localize(self.__datetime)

        return self.__datetime

    def __from_timestamp(self):
        from ..type._integer import Integer
        from ..type._realnumber import RealNumber

        conv_error = TypeConversionError(
            "timestamp is out of the range of values supported by the platform"
        )

        timestamp = Integer(self._value, strict_level=1).try_convert()
        if timestamp:
            try:
                self.__datetime = datetime.fromtimestamp(timestamp, self.__timezone)
            except (ValueError, OSError, OverflowError):
                raise conv_error

            return self.__datetime

        timestamp = RealNumber(self._value, strict_level=1).try_convert()
        if timestamp:
            try:
                self.__datetime = datetime.fromtimestamp(int(timestamp), self.__timezone).replace(
                    microsecond=int((timestamp - int(timestamp)) * 1000000)
                )
            except (ValueError, OSError, OverflowError):
                raise conv_error

            return self.__datetime

        return None

    def __from_datetime_string(self):
        import dateutil.parser
        import pytz

        self.__validate_datetime_string()

        try:
            self.__datetime = dateutil.parser.parse(self._value)
        except (AttributeError, ValueError, OverflowError):
            if self._params.get(ParamKey.STRIP_ANSI_ESCAPE, DefaultValue.STRIP_ANSI_ESCAPE):
                try:
                    self.__datetime = dateutil.parser.parse(strip_ansi_escape(self._value))
                except (AttributeError, ValueError, OverflowError):
                    pass

        if self.__datetime is None:
            raise TypeConversionError(f"failed to parse as a datetime: type={type(self._value)}")

        if self.__timezone:
            pytz_timezone = self.__timezone
        else:
            try:
                dst_timezone_name = self.__get_dst_timezone_name(self.__get_timedelta_sec())
            except (AttributeError, KeyError):
                return self.__datetime

            pytz_timezone = pytz.timezone(dst_timezone_name)

        self.__datetime = self.__datetime.replace(tzinfo=None)
        self.__datetime = pytz_timezone.localize(self.__datetime)

        return self.__datetime

    def __get_timedelta_sec(self):
        dt = self.__datetime.utcoffset()

        return int(
            dt.days * self.__DAYS_TO_SECONDS_COEF
            + float(dt.seconds)
            + dt.microseconds / self.__MICROSECONDS_TO_SECONDS_COEF
        )

    def __get_dst_timezone_name(self, offset):
        return self.__COMMON_DST_TIMEZONE_TABLE[offset]

    def __validate_datetime_string(self):
        """
        This will require validating version string (such as "3.3.5").
        A version string could be converted to a datetime value if this
        validation is not executed.
        """

        from packaging.version import InvalidVersion, Version

        try:
            try:
                Version(self._value)
                raise TypeConversionError(
                    f"invalid datetime string: version string found {self._value}"
                )
            except InvalidVersion:
                pass
        except TypeError:
            raise TypeConversionError(f"invalid datetime string: type={type(self._value)}")



import copy
import warnings
from typing import Any, Optional, Tuple, cast

import dataproperty
import typepy
from mbstrdecoder import MultiByteStrDecoder
from pathvalidate import replace_symbol

from ...error import EmptyTableDataError
from ...sanitizer import sanitize_python_var_name
from ...style import Align, FontStyle, FontWeight, HtmlStyler, Style, StylerInterface, VerticalAlign
from .._common import import_error_msg_template
from .._table_writer import AbstractTableWriter
from ._css import CssTableWriter
from ._text_writer import TextTableWriter


def _get_tags_module() -> Tuple:
    try:
        from dominate import tags
        from dominate.util import raw

        return tags, raw
    except ImportError:
        warnings.warn(import_error_msg_template.format("html"))
        raise


class HtmlTableWriter(TextTableWriter):
    """
    A table writer class for HTML format.

        :Example:
            :ref:`example-html-table-writer`
    """

    FORMAT_NAME = "html"

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    @property
    def support_split_write(self) -> bool:
        return False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.is_padding = False
        self.indent_string = kwargs.get("indent_string", "    ")

        self._dp_extractor.preprocessor.line_break_repl = "<br>"
        self._dp_extractor.preprocessor.is_escape_html_tag = False
        self._quoting_flags = copy.deepcopy(dataproperty.NOT_QUOTING_FLAGS)
        self._table_tag: Any = None

        self.enable_ansi_escape = False

    def write_table(self, **kwargs) -> None:
        """
        |write_table| with HTML table format.

        Args:
            write_css (bool):
                If |True|, write CSS corresponding to the specified styles,
                instead of attributes of HTML tags.

        Example:
            :ref:`example-html-table-writer`

        .. note::
            - |None| values will be replaced with an empty value
        """

        tags, raw = _get_tags_module()
        write_css = kwargs.get("write_css", False)

        with self._logger:
            try:
                self._verify_property()
            except EmptyTableDataError:
                self._logger.logger.debug("no tabular data found")
                return

            self._preprocess()

            css_class = None

            if write_css:
                css_class = kwargs.get(
                    "css_class",
                    "{}-css".format(replace_symbol(self.table_name, replacement_text="-")),
                )

                css_writer = CssTableWriter(table_name=css_class)
                css_writer.from_writer(self, is_overwrite_table_name=False)
                css_writer.write_table(write_style_tag=True)

            if typepy.is_not_null_string(self.table_name):
                if css_class:
                    self._table_tag = tags.table(
                        id=sanitize_python_var_name(self.table_name), class_name=css_class
                    )
                else:
                    self._table_tag = tags.table(id=sanitize_python_var_name(self.table_name))
                self._table_tag += tags.caption(MultiByteStrDecoder(self.table_name).unicode_str)
            else:
                self._table_tag = tags.table()

            try:
                self._write_header()
            except ValueError:
                pass

            self._write_body(not write_css)

    def _write_header(self) -> None:
        tags, raw = _get_tags_module()

        if not self.is_write_header:
            return

        if typepy.is_empty_sequence(self._table_headers):
            raise ValueError("headers is empty")

        tr_tag = tags.tr()
        for header in self._table_headers:
            tr_tag += tags.th(raw(MultiByteStrDecoder(header).unicode_str))

        thead_tag = tags.thead()
        thead_tag += tr_tag

        self._table_tag += thead_tag

    def _write_body(self, write_attr: bool) -> None:
        tags, raw = _get_tags_module()
        tbody_tag = tags.tbody()

        for row_idx, (values, value_dp_list) in enumerate(
            zip(self._table_value_matrix, self._table_value_dp_matrix)
        ):
            tr_tag = tags.tr()
            for value, value_dp, column_dp in zip(values, value_dp_list, self._column_dp_list):
                td_tag = tags.td(raw(MultiByteStrDecoder(value).unicode_str))

                default_style = self._get_col_style(column_dp.column_index)
                style = self._fetch_style_from_filter(row_idx, column_dp, value_dp, default_style)

                if write_attr:
                    if style.align == Align.AUTO:
                        td_tag["align"] = value_dp.align.align_string
                    else:
                        td_tag["align"] = style.align.align_string

                    if style.vertical_align != VerticalAlign.BASELINE:
                        td_tag["valign"] = style.vertical_align.align_str

                    style_tag = self.__make_style_tag(style=style)
                    if style_tag:
                        td_tag["style"] = style_tag

                tr_tag += td_tag
            tbody_tag += tr_tag

        self._table_tag += tbody_tag
        self._write_line(self._table_tag.render(indent=self.indent_string))

    def __make_style_tag(self, style: Style) -> Optional[str]:
        styles = []  # List[str]

        if self._styler.get_font_size(style):
            styles.append(cast(str, self._styler.get_font_size(style)))
        if style.font_weight == FontWeight.BOLD:
            styles.append("font-weight:bold")
        if style.font_style == FontStyle.ITALIC:
            styles.append("font-style:italic")

        if not styles:
            return None

        return "; ".join(styles)

    def _create_styler(self, writer: AbstractTableWriter) -> StylerInterface:
        return HtmlStyler(writer)



# High level events that make up HTTP/1.1 conversations. Loosely inspired by
# the corresponding events in hyper-h2:
#
#     http://python-hyper.org/h2/en/stable/api.html#events
#
# Don't subclass these. Stuff will break.

import re

from . import _headers
from ._abnf import request_target
from ._util import bytesify, LocalProtocolError, validate

# Everything in __all__ gets re-exported as part of the h11 public API.
__all__ = [
    "Request",
    "InformationalResponse",
    "Response",
    "Data",
    "EndOfMessage",
    "ConnectionClosed",
]

request_target_re = re.compile(request_target.encode("ascii"))


class _EventBundle:
    _fields = []
    _defaults = {}

    def __init__(self, **kwargs):
        _parsed = kwargs.pop("_parsed", False)
        allowed = set(self._fields)
        for kwarg in kwargs:
            if kwarg not in allowed:
                raise TypeError(
                    "unrecognized kwarg {} for {}".format(
                        kwarg, self.__class__.__name__
                    )
                )
        required = allowed.difference(self._defaults)
        for field in required:
            if field not in kwargs:
                raise TypeError(
                    "missing required kwarg {} for {}".format(
                        field, self.__class__.__name__
                    )
                )
        self.__dict__.update(self._defaults)
        self.__dict__.update(kwargs)

        # Special handling for some fields

        if "headers" in self.__dict__:
            self.headers = _headers.normalize_and_validate(
                self.headers, _parsed=_parsed
            )

        if not _parsed:
            for field in ["method", "target", "http_version", "reason"]:
                if field in self.__dict__:
                    self.__dict__[field] = bytesify(self.__dict__[field])

            if "status_code" in self.__dict__:
                if not isinstance(self.status_code, int):
                    raise LocalProtocolError("status code must be integer")
                # Because IntEnum objects are instances of int, but aren't
                # duck-compatible (sigh), see gh-72.
                self.status_code = int(self.status_code)

        self._validate()

    def _validate(self):
        pass

    def __repr__(self):
        name = self.__class__.__name__
        kwarg_strs = [
            "{}={}".format(field, self.__dict__[field]) for field in self._fields
        ]
        kwarg_str = ", ".join(kwarg_strs)
        return "{}({})".format(name, kwarg_str)

    # Useful for tests
    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    # This is an unhashable type.
    __hash__ = None


class Request(_EventBundle):
    """The beginning of an HTTP request.

    Fields:

    .. attribute:: method

       An HTTP method, e.g. ``b"GET"`` or ``b"POST"``. Always a byte
       string. :term:`Bytes-like objects <bytes-like object>` and native
       strings containing only ascii characters will be automatically
       converted to byte strings.

    .. attribute:: target

       The target of an HTTP request, e.g. ``b"/index.html"``, or one of the
       more exotic formats described in `RFC 7320, section 5.3
       <https://tools.ietf.org/html/rfc7230#section-5.3>`_. Always a byte
       string. :term:`Bytes-like objects <bytes-like object>` and native
       strings containing only ascii characters will be automatically
       converted to byte strings.

    .. attribute:: headers

       Request headers, represented as a list of (name, value) pairs. See
       :ref:`the header normalization rules <headers-format>` for details.

    .. attribute:: http_version

       The HTTP protocol version, represented as a byte string like
       ``b"1.1"``. See :ref:`the HTTP version normalization rules
       <http_version-format>` for details.

    """

    _fields = ["method", "target", "headers", "http_version"]
    _defaults = {"http_version": b"1.1"}

    def _validate(self):
        # "A server MUST respond with a 400 (Bad Request) status code to any
        # HTTP/1.1 request message that lacks a Host header field and to any
        # request message that contains more than one Host header field or a
        # Host header field with an invalid field-value."
        # -- https://tools.ietf.org/html/rfc7230#section-5.4
        host_count = 0
        for name, value in self.headers:
            if name == b"host":
                host_count += 1
        if self.http_version == b"1.1" and host_count == 0:
            raise LocalProtocolError("Missing mandatory Host: header")
        if host_count > 1:
            raise LocalProtocolError("Found multiple Host: headers")

        validate(request_target_re, self.target, "Illegal target characters")


class _ResponseBase(_EventBundle):
    _fields = ["status_code", "headers", "http_version", "reason"]
    _defaults = {"http_version": b"1.1", "reason": b""}


class InformationalResponse(_ResponseBase):
    """An HTTP informational response.

    Fields:

    .. attribute:: status_code

       The status code of this response, as an integer. For an
       :class:`InformationalResponse`, this is always in the range [100,
       200).

    .. attribute:: headers

       Request headers, represented as a list of (name, value) pairs. See
       :ref:`the header normalization rules <headers-format>` for
       details.

    .. attribute:: http_version

       The HTTP protocol version, represented as a byte string like
       ``b"1.1"``. See :ref:`the HTTP version normalization rules
       <http_version-format>` for details.

    .. attribute:: reason

       The reason phrase of this response, as a byte string. For example:
       ``b"OK"``, or ``b"Not Found"``.

    """

    def _validate(self):
        if not (100 <= self.status_code < 200):
            raise LocalProtocolError(
                "InformationalResponse status_code should be in range "
                "[100, 200), not {}".format(self.status_code)
            )


class Response(_ResponseBase):
    """The beginning of an HTTP response.

    Fields:

    .. attribute:: status_code

       The status code of this response, as an integer. For an
       :class:`Response`, this is always in the range [200,
       600).

    .. attribute:: headers

       Request headers, represented as a list of (name, value) pairs. See
       :ref:`the header normalization rules <headers-format>` for details.

    .. attribute:: http_version

       The HTTP protocol version, represented as a byte string like
       ``b"1.1"``. See :ref:`the HTTP version normalization rules
       <http_version-format>` for details.

    .. attribute:: reason

       The reason phrase of this response, as a byte string. For example:
       ``b"OK"``, or ``b"Not Found"``.

    """

    def _validate(self):
        if not (200 <= self.status_code < 600):
            raise LocalProtocolError(
                "Response status_code should be in range [200, 600), not {}".format(
                    self.status_code
                )
            )


class Data(_EventBundle):
    """Part of an HTTP message body.

    Fields:

    .. attribute:: data

       A :term:`bytes-like object` containing part of a message body. Or, if
       using the ``combine=False`` argument to :meth:`Connection.send`, then
       any object that your socket writing code knows what to do with, and for
       which calling :func:`len` returns the number of bytes that will be
       written -- see :ref:`sendfile` for details.

    .. attribute:: chunk_start

       A marker that indicates whether this data object is from the start of a
       chunked transfer encoding chunk. This field is ignored when when a Data
       event is provided to :meth:`Connection.send`: it is only valid on
       events emitted from :meth:`Connection.next_event`. You probably
       shouldn't use this attribute at all; see
       :ref:`chunk-delimiters-are-bad` for details.

    .. attribute:: chunk_end

       A marker that indicates whether this data object is the last for a
       given chunked transfer encoding chunk. This field is ignored when when
       a Data event is provided to :meth:`Connection.send`: it is only valid
       on events emitted from :meth:`Connection.next_event`. You probably
       shouldn't use this attribute at all; see
       :ref:`chunk-delimiters-are-bad` for details.

    """

    _fields = ["data", "chunk_start", "chunk_end"]
    _defaults = {"chunk_start": False, "chunk_end": False}


# XX FIXME: "A recipient MUST ignore (or consider as an error) any fields that
# are forbidden to be sent in a trailer, since processing them as if they were
# present in the header section might bypass external security filters."
# https://svn.tools.ietf.org/svn/wg/httpbis/specs/rfc7230.html#chunked.trailer.part
# Unfortunately, the list of forbidden fields is long and vague :-/
class EndOfMessage(_EventBundle):
    """The end of an HTTP message.

    Fields:

    .. attribute:: headers

       Default value: ``[]``

       Any trailing headers attached to this message, represented as a list of
       (name, value) pairs. See :ref:`the header normalization rules
       <headers-format>` for details.

       Must be empty unless ``Transfer-Encoding: chunked`` is in use.

    """

    _fields = ["headers"]
    _defaults = {"headers": []}


class ConnectionClosed(_EventBundle):
    """This event indicates that the sender has closed their outgoing
    connection.

    Note that this does not necessarily mean that they can't *receive* further
    data, because TCP connections are composed to two one-way channels which
    can be closed independently. See :ref:`closing` for details.

    No fields.
    """

    pass



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from .._const import StrictLevel
from ._checker import CheckerFactory, TypeCheckerBase, TypeCheckerDelegator
from ._common import isstring


class BoolTypeCheckerStrictLevel0(TypeCheckerBase):
    def is_instance(self):
        return isinstance(self._value, bool)

    def is_valid_after_convert(self, converted_value):
        return isinstance(converted_value, bool)


class BoolTypeCheckerStrictLevel1(BoolTypeCheckerStrictLevel0):
    def is_exclude_instance(self):
        from ..type._integer import Integer

        return Integer(self._value, strict_level=StrictLevel.MAX).is_type()


class BoolTypeCheckerStrictLevel2(BoolTypeCheckerStrictLevel1):
    def is_exclude_instance(self):
        return super().is_exclude_instance() or isstring(self._value)


_factory = CheckerFactory(
    checker_mapping={
        0: BoolTypeCheckerStrictLevel0,
        1: BoolTypeCheckerStrictLevel1,
        2: BoolTypeCheckerStrictLevel2,
    }
)


class BoolTypeChecker(TypeCheckerDelegator):
    def __init__(self, value, strict_level):
        super().__init__(value=value, checker_factory=_factory, strict_level=strict_level)



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import enum


@enum.unique
class Align(enum.Enum):
    AUTO = (1 << 0, "auto")
    LEFT = (1 << 1, "left")
    RIGHT = (1 << 2, "right")
    CENTER = (1 << 3, "center")

    @property
    def align_code(self) -> int:
        return self.__align_code

    @property
    def align_string(self) -> str:
        return self.__align_string

    def __init__(self, code: int, string: str):
        self.__align_code = code
        self.__align_string = string



'''
Utility classes for creating dynamic html documents
'''

__license__ = '''
This file is part of Dominate.

Dominate is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

Dominate is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with Dominate.  If not, see
<http://www.gnu.org/licenses/>.
'''

import re

from .dom_tag import dom_tag

try:
  basestring = basestring
except NameError:
  basestring = str
  unichr = chr


def include(f):
  '''
  includes the contents of a file on disk.
  takes a filename
  '''
  fl = open(f, 'r')
  data = fl.read()
  fl.close()
  return raw(data)


def system(cmd, data=None):
  '''
  pipes the output of a program
  '''
  import subprocess
  s = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
  out, err = s.communicate(data)
  return out.decode('utf8')


def escape(data, quote=True):  # stolen from std lib cgi
  '''
  Escapes special characters into their html entities
  Replace special characters "&", "<" and ">" to HTML-safe sequences.
  If the optional flag quote is true, the quotation mark character (")
  is also translated.

  This is used to escape content that appears in the body of an HTML document
  '''
  data = data.replace("&", "&amp;")  # Must be done first!
  data = data.replace("<", "&lt;")
  data = data.replace(">", "&gt;")
  if quote:
    data = data.replace('"', "&quot;")
  return data


_unescape = {
  'quot': 34,
  'amp':  38,
  'lt':   60,
  'gt':   62,
  'nbsp': 32,
  # more here
  # http://www.w3.org/TR/html4/sgml/entities.html
  'yuml': 255,
}
str_escape = escape


def unescape(data):
  '''
  unescapes html entities. the opposite of escape.
  '''
  cc = re.compile(r'&(?:(?:#(\d+))|([^;]+));')

  result = []
  m = cc.search(data)
  while m:
    result.append(data[0:m.start()])
    d = m.group(1)
    if d:
      d = int(d)
      result.append(unichr(d))
    else:
      d = _unescape.get(m.group(2), ord('?'))
      result.append(unichr(d))

    data = data[m.end():]
    m = cc.search(data)

  result.append(data)
  return ''.join(result)


_reserved = ";/?:@&=+$, "
_replace_map = dict((c, '%%%2X' % ord(c)) for c in _reserved)


def url_escape(data):
  return ''.join(_replace_map.get(c, c) for c in data)


def url_unescape(data):
  return re.sub('%([0-9a-fA-F]{2})',
    lambda m: unichr(int(m.group(1), 16)), data)


class lazy(dom_tag):
  '''
  delays function execution until rendered
  '''
  def __new__(_cls, *args, **kwargs):
    '''
    Need to reset this special method or else
    dom_tag will think it's being used as a dectorator.

    This means lazy() can't be used as a dectorator, but
    thinking about when you might want that just confuses me.
    '''
    return object.__new__(_cls)

  def __init__(self, func, *args, **kwargs):
    super(lazy, self).__init__()
    self.func   = func
    self.args   = args
    self.kwargs = kwargs


  def _render(self, sb, *a, **kw):
    r = self.func(*self.args, **self.kwargs)
    sb.append(str(r))


# TODO rename this to raw?
class text(dom_tag):
  '''
  Just a string. useful for inside context managers
  '''
  is_pretty = False
  is_inline = True

  def __init__(self, _text, escape=True):
    super(text, self).__init__()
    if escape:
      self.text = str_escape(_text)
    else:
      self.text = _text

  def _render(self, sb, *a, **kw):
    sb.append(self.text)
    return sb


def raw(s):
  '''
  Inserts a raw string into the DOM. Unsafe.
  '''
  return text(s, escape=False)



from ._elasticsearch import ElasticsearchWriter
from ._null import NullTableWriter
from ._table_writer import AbstractTableWriter
from .binary import (
    ExcelXlsTableWriter,
    ExcelXlsxTableWriter,
    PandasDataFramePickleWriter,
    SqliteTableWriter,
)
from .text import (
    AsciiDocTableWriter,
    BoldUnicodeTableWriter,
    BorderlessTableWriter,
    CssTableWriter,
    CsvTableWriter,
    HtmlTableWriter,
    JsonLinesTableWriter,
    JsonTableWriter,
    LatexMatrixWriter,
    LatexTableWriter,
    LtsvTableWriter,
    MarkdownTableWriter,
    MediaWikiTableWriter,
    RstCsvTableWriter,
    RstGridTableWriter,
    RstSimpleTableWriter,
    SpaceAlignedTableWriter,
    TomlTableWriter,
    TsvTableWriter,
    UnicodeTableWriter,
    YamlTableWriter,
)
from .text.sourcecode import (
    JavaScriptTableWriter,
    NumpyTableWriter,
    PandasDataFrameWriter,
    PythonCodeTableWriter,
)



import abc

from .._table_writer import AbstractTableWriter


class BinaryWriterInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def is_opened(self) -> bool:  # pragma: no cover
        pass

    @abc.abstractmethod
    def open(self, file_path: str) -> None:  # pragma: no cover
        """
        Open a file for output stream.

        Args:
            file_path (str): path to the file.
        """


class AbstractBinaryTableWriter(AbstractTableWriter, BinaryWriterInterface):
    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, value) -> None:
        raise RuntimeError(
            "cannot assign a stream to binary format writers. use open method instead."
        )

    @property
    def support_split_write(self) -> bool:
        return True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.table_name = kwargs.get("table_name", "")

        self._stream = None

    def __del__(self) -> None:
        self.close()

    def is_opened(self) -> bool:
        return self.stream is not None

    def dumps(self) -> str:
        raise NotImplementedError("binary format writers did not support dumps method")

    def _verify_stream(self) -> None:
        if self.stream is None:
            raise OSError("null output stream. required to open(file_path) first.")

    def _write_value_row_separator(self) -> None:
        pass



import select
import socket
import sys
import typing


def is_socket_readable(sock: typing.Optional[socket.socket]) -> bool:
    """
    Return whether a socket, as identifed by its file descriptor, is readable.
    "A socket is readable" means that the read buffer isn't empty, i.e. that calling
    .recv() on it would immediately return some data.
    """
    # NOTE: we want check for readability without actually attempting to read, because
    # we don't want to block forever if it's not readable.

    # In the case that the socket no longer exists, or cannot return a file
    # descriptor, we treat it as being readable, as if it the next read operation
    # on it is ready to return the terminating `b""`.
    sock_fd = None if sock is None else sock.fileno()
    if sock_fd is None or sock_fd < 0:  # pragma: nocover
        return True

    # The implementation below was stolen from:
    # https://github.com/python-trio/trio/blob/20ee2b1b7376db637435d80e266212a35837ddcc/trio/_socket.py#L471-L478
    # See also: https://github.com/encode/httpcore/pull/193#issuecomment-703129316

    # Use select.select on Windows, and when poll is unavailable and select.poll
    # everywhere else. (E.g. When eventlet is in use. See #327)
    if (
        sys.platform == "win32" or getattr(select, "poll", None) is None
    ):  # pragma: nocover
        rready, _, _ = select.select([sock_fd], [], [], 0)
        return bool(rready)
    p = select.poll()
    p.register(sock_fd, select.POLLIN)
    return bool(p.poll(0))



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from decimal import Decimal


class DefaultValue:
    FLOAT_TYPE = Decimal
    STRIP_ANSI_ESCAPE = True


class StrictLevel:
    MIN = 0
    MAX = 100


class ParamKey:
    STRIP_ANSI_ESCAPE = "strip_ansi_escape"
    TIMEZONE = "timezone"



# A highish-level implementation of the HTTP/1.1 wire protocol (RFC 7230),
# containing no networking code at all, loosely modelled on hyper-h2's generic
# implementation of HTTP/2 (and in particular the h2.connection.H2Connection
# class). There's still a bunch of subtle details you need to get right if you
# want to make this actually useful, because it doesn't implement all the
# semantics to check that what you're asking to write to the wire is sensible,
# but at least it gets you out of dealing with the wire itself.

from ._connection import *
from ._events import *
from ._state import *
from ._util import LocalProtocolError, ProtocolError, RemoteProtocolError
from ._version import __version__

PRODUCT_ID = "python-h11/" + __version__


__all__ = ["ProtocolError", "LocalProtocolError", "RemoteProtocolError"]
__all__ += _events.__all__
__all__ += _connection.__all__
__all__ += _state.__all__



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from ._checker import CheckerFactory, TypeCheckerBase, TypeCheckerDelegator
from ._common import isstring


class IpAddressTypeCheckerStrictLevel0(TypeCheckerBase):
    def is_instance(self):
        return self._is_ipaddress(self._value)

    def is_valid_after_convert(self, converted_value):
        return self._is_ipaddress(converted_value)

    @staticmethod
    def _is_ipaddress(value):
        import ipaddress

        return isinstance(value, (ipaddress.IPv4Address, ipaddress.IPv6Address))


class IpAddressTypeCheckerStrictLevel1(IpAddressTypeCheckerStrictLevel0):
    def is_exclude_instance(self):
        return isstring(self._value) or super().is_exclude_instance()


_factory = CheckerFactory(
    checker_mapping={0: IpAddressTypeCheckerStrictLevel0, 1: IpAddressTypeCheckerStrictLevel1}
)


class IpAddressTypeChecker(TypeCheckerDelegator):
    def __init__(self, value, strict_level):
        super().__init__(value=value, checker_factory=_factory, strict_level=strict_level)



from collections import OrderedDict, deque
from dataclasses import dataclass, field
from types import TracebackType
from typing import Deque, Generic, List, NamedTuple, Optional, Type, TypeVar

from .. import (
    BrokenResourceError, ClosedResourceError, EndOfStream, WouldBlock, get_cancelled_exc_class)
from .._core._compat import DeprecatedAwaitable
from ..abc import Event, ObjectReceiveStream, ObjectSendStream
from ..lowlevel import checkpoint

T_Item = TypeVar('T_Item')


class MemoryObjectStreamStatistics(NamedTuple):
    current_buffer_used: int  #: number of items stored in the buffer
    #: maximum number of items that can be stored on this stream (or :data:`math.inf`)
    max_buffer_size: float
    open_send_streams: int  #: number of unclosed clones of the send stream
    open_receive_streams: int  #: number of unclosed clones of the receive stream
    tasks_waiting_send: int  #: number of tasks blocked on :meth:`MemoryObjectSendStream.send`
    #: number of tasks blocked on :meth:`MemoryObjectReceiveStream.receive`
    tasks_waiting_receive: int


@dataclass(eq=False)
class MemoryObjectStreamState(Generic[T_Item]):
    max_buffer_size: float = field()
    buffer: Deque[T_Item] = field(init=False, default_factory=deque)
    open_send_channels: int = field(init=False, default=0)
    open_receive_channels: int = field(init=False, default=0)
    waiting_receivers: 'OrderedDict[Event, List[T_Item]]' = field(init=False,
                                                                  default_factory=OrderedDict)
    waiting_senders: 'OrderedDict[Event, T_Item]' = field(init=False, default_factory=OrderedDict)

    def statistics(self) -> MemoryObjectStreamStatistics:
        return MemoryObjectStreamStatistics(
            len(self.buffer), self.max_buffer_size, self.open_send_channels,
            self.open_receive_channels, len(self.waiting_senders), len(self.waiting_receivers))


@dataclass(eq=False)
class MemoryObjectReceiveStream(Generic[T_Item], ObjectReceiveStream[T_Item]):
    _state: MemoryObjectStreamState[T_Item]
    _closed: bool = field(init=False, default=False)

    def __post_init__(self) -> None:
        self._state.open_receive_channels += 1

    def receive_nowait(self) -> T_Item:
        """
        Receive the next item if it can be done without waiting.

        :return: the received item
        :raises ~anyio.ClosedResourceError: if this send stream has been closed
        :raises ~anyio.EndOfStream: if the buffer is empty and this stream has been
            closed from the sending end
        :raises ~anyio.WouldBlock: if there are no items in the buffer and no tasks
            waiting to send

        """
        if self._closed:
            raise ClosedResourceError

        if self._state.waiting_senders:
            # Get the item from the next sender
            send_event, item = self._state.waiting_senders.popitem(last=False)
            self._state.buffer.append(item)
            send_event.set()

        if self._state.buffer:
            return self._state.buffer.popleft()
        elif not self._state.open_send_channels:
            raise EndOfStream

        raise WouldBlock

    async def receive(self) -> T_Item:
        await checkpoint()
        try:
            return self.receive_nowait()
        except WouldBlock:
            # Add ourselves in the queue
            receive_event = Event()
            container: List[T_Item] = []
            self._state.waiting_receivers[receive_event] = container

            try:
                await receive_event.wait()
            except get_cancelled_exc_class():
                # Ignore the immediate cancellation if we already received an item, so as not to
                # lose it
                if not container:
                    raise
            finally:
                self._state.waiting_receivers.pop(receive_event, None)

            if container:
                return container[0]
            else:
                raise EndOfStream

    def clone(self) -> 'MemoryObjectReceiveStream':
        """
        Create a clone of this receive stream.

        Each clone can be closed separately. Only when all clones have been closed will the
        receiving end of the memory stream be considered closed by the sending ends.

        :return: the cloned stream

        """
        if self._closed:
            raise ClosedResourceError

        return MemoryObjectReceiveStream(_state=self._state)

    def close(self) -> None:
        """
        Close the stream.

        This works the exact same way as :meth:`aclose`, but is provided as a special case for the
        benefit of synchronous callbacks.

        """
        if not self._closed:
            self._closed = True
            self._state.open_receive_channels -= 1
            if self._state.open_receive_channels == 0:
                send_events = list(self._state.waiting_senders.keys())
                for event in send_events:
                    event.set()

    async def aclose(self) -> None:
        self.close()

    def statistics(self) -> MemoryObjectStreamStatistics:
        """
        Return statistics about the current state of this stream.

        .. versionadded:: 3.0
        """
        return self._state.statistics()

    def __enter__(self) -> 'MemoryObjectReceiveStream[T_Item]':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        self.close()


@dataclass(eq=False)
class MemoryObjectSendStream(Generic[T_Item], ObjectSendStream[T_Item]):
    _state: MemoryObjectStreamState[T_Item]
    _closed: bool = field(init=False, default=False)

    def __post_init__(self) -> None:
        self._state.open_send_channels += 1

    def send_nowait(self, item: T_Item) -> DeprecatedAwaitable:
        """
        Send an item immediately if it can be done without waiting.

        :param item: the item to send
        :raises ~anyio.ClosedResourceError: if this send stream has been closed
        :raises ~anyio.BrokenResourceError: if the stream has been closed from the
            receiving end
        :raises ~anyio.WouldBlock: if the buffer is full and there are no tasks waiting
            to receive

        """
        if self._closed:
            raise ClosedResourceError
        if not self._state.open_receive_channels:
            raise BrokenResourceError

        if self._state.waiting_receivers:
            receive_event, container = self._state.waiting_receivers.popitem(last=False)
            container.append(item)
            receive_event.set()
        elif len(self._state.buffer) < self._state.max_buffer_size:
            self._state.buffer.append(item)
        else:
            raise WouldBlock

        return DeprecatedAwaitable(self.send_nowait)

    async def send(self, item: T_Item) -> None:
        await checkpoint()
        try:
            self.send_nowait(item)
        except WouldBlock:
            # Wait until there's someone on the receiving end
            send_event = Event()
            self._state.waiting_senders[send_event] = item
            try:
                await send_event.wait()
            except BaseException:
                self._state.waiting_senders.pop(send_event, None)  # type: ignore[arg-type]
                raise

            if self._state.waiting_senders.pop(send_event, None):  # type: ignore[arg-type]
                raise BrokenResourceError

    def clone(self) -> 'MemoryObjectSendStream':
        """
        Create a clone of this send stream.

        Each clone can be closed separately. Only when all clones have been closed will the
        sending end of the memory stream be considered closed by the receiving ends.

        :return: the cloned stream

        """
        if self._closed:
            raise ClosedResourceError

        return MemoryObjectSendStream(_state=self._state)

    def close(self) -> None:
        """
        Close the stream.

        This works the exact same way as :meth:`aclose`, but is provided as a special case for the
        benefit of synchronous callbacks.

        """
        if not self._closed:
            self._closed = True
            self._state.open_send_channels -= 1
            if self._state.open_send_channels == 0:
                receive_events = list(self._state.waiting_receivers.keys())
                self._state.waiting_receivers.clear()
                for event in receive_events:
                    event.set()

    async def aclose(self) -> None:
        self.close()

    def statistics(self) -> MemoryObjectStreamStatistics:
        """
        Return statistics about the current state of this stream.

        .. versionadded:: 3.0
        """
        return self._state.statistics()

    def __enter__(self) -> 'MemoryObjectSendStream[T_Item]':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        self.close()



import abc
import copy
import warnings
from typing import Any, Dict, Optional, cast

import dataproperty
import typepy
from dataproperty import DataProperty
from tabledata import TableData
from typepy import Integer

from .._common import import_error_msg_template
from ._excel_workbook import ExcelWorkbookInterface, ExcelWorkbookXls, ExcelWorkbookXlsx
from ._interface import AbstractBinaryTableWriter


class ExcelTableWriter(AbstractBinaryTableWriter, metaclass=abc.ABCMeta):
    """
    An abstract class of a table writer for Excel file format.
    """

    FORMAT_NAME = "excel"

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    @property
    def workbook(self) -> Optional[ExcelWorkbookInterface]:
        return self._workbook

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._workbook: Optional[ExcelWorkbookInterface] = None

        self._dp_extractor.type_value_map = {
            typepy.Typecode.INFINITY: "Inf",
            typepy.Typecode.NAN: "NaN",
        }

        self._first_header_row = 0
        self._last_header_row = self.first_header_row
        self._first_data_row = self.last_header_row + 1
        self._first_data_col = 0
        self._last_data_row: Optional[int] = None
        self._last_data_col: Optional[int] = None

        self._current_data_row = self._first_data_row

        self._quoting_flags = copy.deepcopy(dataproperty.NOT_QUOTING_FLAGS)
        self._quoting_flags[typepy.Typecode.DATETIME] = True

    @property
    def first_header_row(self) -> int:
        """
        :return: Index of the first row of the header.
        :rtype: int

        .. note:: |excel_attr|
        """

        return self._first_header_row

    @property
    def last_header_row(self) -> int:
        """
        :return: Index of the last row of the header.
        :rtype: int

        .. note:: |excel_attr|
        """

        return self._last_header_row

    @property
    def first_data_row(self) -> int:
        """
        :return: Index of the first row of the data (table body).
        :rtype: int

        .. note:: |excel_attr|
        """

        return self._first_data_row

    @property
    def last_data_row(self) -> Optional[int]:
        """
        :return: Index of the last row of the data (table body).
        :rtype: int

        .. note:: |excel_attr|
        """

        return self._last_data_row

    @property
    def first_data_col(self) -> int:
        """
        :return: Index of the first column of the table.
        :rtype: int

        .. note:: |excel_attr|
        """

        return self._first_data_col

    @property
    def last_data_col(self) -> Optional[int]:
        """
        :return: Index of the last column of the table.
        :rtype: int

        .. note:: |excel_attr|
        """

        return self._last_data_col

    def is_opened(self) -> bool:
        return self.workbook is not None

    def open(self, file_path: str) -> None:
        """
        Open an Excel workbook file.

        :param str file_path: Excel workbook file path to open.
        """

        if self.workbook and self.workbook.file_path == file_path:
            self._logger.logger.debug(f"workbook already opened: {self.workbook.file_path}")
            return

        self.close()
        self._open(file_path)

    @abc.abstractmethod
    def _open(self, workbook_path: str) -> None:  # pragma: no cover
        pass

    def close(self) -> None:
        """
        Close the current workbook.
        """

        if self.is_opened():
            self.workbook.close()  # type: ignore
            self._workbook = None

    def from_tabledata(self, value: TableData, is_overwrite_table_name: bool = True) -> None:
        """
        Set following attributes from |TableData|

        - :py:attr:`~.table_name`.
        - :py:attr:`~.headers`.
        - :py:attr:`~.value_matrix`.

        And create worksheet named from :py:attr:`~.table_name` ABC
        if not existed yet.

        :param tabledata.TableData value: Input table data.
        """

        super().from_tabledata(value)

        if self.is_opened():
            self.make_worksheet(self.table_name)

    def make_worksheet(self, sheet_name: Optional[str] = None) -> None:
        """Make a worksheet to the current workbook.

        Args:
            sheet_name (str):
                Name of the worksheet to create. The name will be automatically generated
                (like ``"Sheet1"``) if the ``sheet_name`` is empty.
        """

        if sheet_name is None:
            sheet_name = self.table_name
        if not sheet_name:
            sheet_name = ""

        self._stream = self.workbook.add_worksheet(sheet_name)  # type: ignore
        self._current_data_row = self._first_data_row

    def dump(self, output: str, close_after_write: bool = True, **kwargs) -> None:
        """Write a worksheet to the current workbook.

        Args:
            output (str):
                Path to the workbook file to write.
            close_after_write (bool, optional):
                Close the workbook after write.
                Defaults to |True|.
        """

        self.open(output)
        try:
            self.make_worksheet(self.table_name)
            self.write_table(**kwargs)
        finally:
            if close_after_write:
                self.close()

    @abc.abstractmethod
    def _write_header(self) -> None:
        pass

    @abc.abstractmethod
    def _write_cell(self, row: int, col: int, value_dp: DataProperty) -> None:
        pass

    def _write_table(self, **kwargs) -> None:
        self._preprocess_table_dp()
        self._preprocess_table_property()
        self._write_header()
        self._write_value_matrix()
        self._postprocess()

    def _write_value_matrix(self) -> None:
        for value_dp_list in self._table_value_dp_matrix:
            for col_idx, value_dp in enumerate(value_dp_list):
                self._write_cell(self._current_data_row, col_idx, value_dp)

            self._current_data_row += 1

    def _get_last_column(self) -> int:
        if typepy.is_not_empty_sequence(self.headers):
            return len(self.headers) - 1

        if typepy.is_not_empty_sequence(self.value_matrix):
            return len(self.value_matrix[0]) - 1

        raise ValueError("data not found")

    def _postprocess(self) -> None:
        self._last_data_row = self._current_data_row
        self._last_data_col = self._get_last_column()


class ExcelXlsTableWriter(ExcelTableWriter):
    """
    A table writer class for Excel file format: ``.xls`` (older or equal to Office 2003).

    ``xlwt`` package required to use this class.

    .. py:method:: write_table()

        Write a table to the current opened worksheet.

        :raises IOError: If failed to write data to the worksheet.

        .. note::
            Specific values in the tabular data are converted when writing:

            - |None|: written as an empty string
            - |inf|: written as ``Inf``
            - |nan|: written as ``NaN``
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.__col_style_table: Dict[int, Any] = {}

    def _open(self, workbook_path: str) -> None:
        self._workbook = ExcelWorkbookXls(workbook_path)

    def _write_header(self) -> None:
        if not self.is_write_header or typepy.is_empty_sequence(self.headers):
            return

        for col, value in enumerate(self.headers):
            self.stream.write(self.first_header_row, col, value)

    def _write_cell(self, row: int, col: int, value_dp: DataProperty) -> None:
        if value_dp.typecode in [typepy.Typecode.REAL_NUMBER]:
            try:
                cell_style = self.__get_cell_style(col)
            except ValueError:
                pass
            else:
                self.stream.write(row, col, value_dp.data, cell_style)
                return

        self.stream.write(row, col, value_dp.data)

    def _postprocess(self) -> None:
        super()._postprocess()

        self.__col_style_table = {}

    def __get_cell_style(self, col: int):
        try:
            import xlwt
        except ImportError:
            warnings.warn(import_error_msg_template.format("excel"))
            raise

        if col in self.__col_style_table:
            return self.__col_style_table.get(col)

        try:
            col_dp = self._column_dp_list[col]
        except KeyError:
            return {}

        if col_dp.typecode not in [typepy.Typecode.REAL_NUMBER]:
            raise ValueError()

        if not Integer(col_dp.minmax_decimal_places.max_value).is_type():
            raise ValueError()

        float_digit = col_dp.minmax_decimal_places.max_value
        if float_digit <= 0:
            raise ValueError()

        num_format_str = "#,{:s}0.{:s}".format("#" * int(float_digit), "0" * int(float_digit))
        cell_style = xlwt.easyxf(num_format_str=num_format_str)
        self.__col_style_table[col] = cell_style

        return cell_style


class ExcelXlsxTableWriter(ExcelTableWriter):
    """
    A table writer class for Excel file format: ``.xlsx`` (newer or equal to Office 2007).

    .. py:method:: write_table()

        Write a table to the current opened worksheet.

        :raises IOError: If failed to write data to the worksheet.
        :Examples:
            :ref:`example-excel-table-writer`

        .. note::
            Specific values in the tabular data are converted when writing:

            - |None|: written as an empty string
            - |inf|: written as ``Inf``
            - |nan|: written as ``NaN``
    """

    MAX_CELL_WIDTH = 60

    class TableFormat:
        HEADER = "header"
        CELL = "cell"
        NAN = "nan"

    class Default:
        FONT_NAME = "MS Gothic"
        FONT_SIZE = 9

        CELL_FORMAT = {
            "font_name": FONT_NAME,
            "font_size": FONT_SIZE,
            "align": "top",
            "text_wrap": True,
            "top": 1,
            "left": 1,
            "bottom": 1,
            "right": 1,
        }
        HEADER_FORMAT = {
            "font_name": FONT_NAME,
            "font_size": FONT_SIZE,
            "bg_color": "#DFDFFF",
            "bold": True,
            "left": 1,
            "right": 1,
        }
        NAN_FORMAT = {
            "font_name": FONT_NAME,
            "font_size": FONT_SIZE,
            "font_color": "silver",
            "top": 1,
            "left": 1,
            "bottom": 1,
            "right": 1,
        }

    @property
    def __nan_format_property(self) -> Dict:
        return self.format_table.get(self.TableFormat.NAN, self.default_format)

    @property
    def __cell_format_property(self) -> Dict:
        return self.format_table.get(self.TableFormat.CELL, self.default_format)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.default_format = self.Default.CELL_FORMAT
        self.format_table = {
            self.TableFormat.CELL: self.Default.CELL_FORMAT,
            self.TableFormat.HEADER: self.Default.HEADER_FORMAT,
            self.TableFormat.NAN: self.Default.NAN_FORMAT,
        }

        self.__col_cell_format_cache: Dict[int, Any] = {}
        self.__col_numprops_table: Dict[int, Dict] = {}

    def _open(self, workbook_path: str) -> None:
        self._workbook = ExcelWorkbookXlsx(workbook_path)

    def _write_header(self) -> None:
        if not self.is_write_header or typepy.is_empty_sequence(self.headers):
            return

        header_format_props = self.format_table.get(self.TableFormat.HEADER, self.default_format)
        header_format = self.__add_format(header_format_props)

        self.stream.write_row(
            row=self.first_header_row, col=0, data=self.headers, cell_format=header_format
        )
        for row in range(self.first_header_row, self.last_header_row):
            self.stream.write_row(
                row=row, col=0, data=[""] * len(self.headers), cell_format=header_format
            )

    def _write_cell(self, row: int, col: int, value_dp: DataProperty) -> None:
        base_props = dict(self.__cell_format_property)
        format_key = f"{col:d}_{value_dp.typecode.name:s}"

        if value_dp.typecode in [typepy.Typecode.INTEGER, typepy.Typecode.REAL_NUMBER]:
            num_props = self.__get_number_property(col)
            base_props.update(num_props)
            cell_format = self.__get_cell_format(format_key, base_props)

            try:
                self.stream.write_number(row, col, float(value_dp.data), cell_format)
                return
            except TypeError:
                pass

        if value_dp.typecode is typepy.Typecode.NAN:
            base_props = dict(self.__nan_format_property)

        cell_format = self.__get_cell_format(format_key, base_props)
        self.stream.write(row, col, value_dp.data, cell_format)

    def __get_number_property(self, col: int) -> Dict:
        if col in self.__col_numprops_table:
            return cast(Dict, self.__col_numprops_table.get(col))

        try:
            col_dp = self._column_dp_list[col]
        except KeyError:
            return {}

        if col_dp.typecode not in [typepy.Typecode.INTEGER, typepy.Typecode.REAL_NUMBER]:
            return {}

        num_props = {}
        if Integer(col_dp.minmax_decimal_places.max_value).is_type():
            float_digit = col_dp.minmax_decimal_places.max_value
            if float_digit > 0:
                num_props = {"num_format": "0.{:s}".format("0" * int(float_digit))}

        self.__col_numprops_table[col] = num_props

        return num_props

    def __get_cell_format(self, format_key, cell_props) -> Dict:
        cell_format = self.__col_cell_format_cache.get(format_key)
        if cell_format is not None:
            return cell_format

        # cache miss
        cell_format = self.__add_format(cell_props)
        self.__col_cell_format_cache[format_key] = cell_format

        return cell_format

    def __add_format(self, dict_property):
        return self.workbook.workbook.add_format(dict_property)

    def __set_cell_width(self):
        font_size = self.__cell_format_property.get("font_size")

        if not Integer(font_size).is_type():
            return

        for col_idx, col_dp in enumerate(self._column_dp_list):
            width = min(col_dp.ascii_char_width, self.MAX_CELL_WIDTH) * (font_size / 10.0) + 2
            self.stream.set_column(col_idx, col_idx, width=width)

    def _preprocess_table_property(self) -> None:
        super()._preprocess_table_property()

        self.__set_cell_width()

    def _postprocess(self) -> None:
        super()._postprocess()

        self.stream.autofilter(
            self.last_header_row, self.first_data_col, self.last_data_row, self.last_data_col
        )
        self.stream.freeze_panes(self.first_data_row, self.first_data_col)

        self.__col_cell_format_cache = {}
        self.__col_numprops_table = {}



import copy
from typing import Any, List, Sequence

import dataproperty as dp
import typepy
from dataproperty import ColumnDataProperty, DataProperty, LineBreakHandling
from mbstrdecoder import MultiByteStrDecoder

from ...style import Align, FontStyle, FontWeight, Style, StylerInterface, TextStyler
from ...style._styler import get_align_char
from .._table_writer import AbstractTableWriter
from ._text_writer import TextTableWriter


class AsciiDocStyler(TextStyler):
    def apply(self, value: Any, style: Style) -> str:
        value = super().apply(value, style)
        if not value:
            return value

        try:
            fg_color = style.fg_color.name.lower()  # type: ignore
        except AttributeError:
            fg_color = None

        try:
            bg_color = style.bg_color.name.lower()  # type: ignore
        except AttributeError:
            bg_color = None

        if fg_color and bg_color:
            value = f"[{fg_color} {bg_color}-background]##{value}##"
        elif fg_color:
            value = f"[{fg_color}]##{value}##"
        elif bg_color:
            value = f"[{bg_color}-background]##{value}##"

        if style.font_weight == FontWeight.BOLD:
            value = f"*{value}*"

        if style.font_style == FontStyle.ITALIC:
            value = f"_{value}_"

        return value


class AsciiDocTableWriter(TextTableWriter):
    """
    A table writer class for `AsciiDoc <https://asciidoc.org/>`__ format.
    """

    FORMAT_NAME = "asciidoc"

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    @property
    def support_split_write(self) -> bool:
        return True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.column_delimiter = "\n"

        self.is_padding = False
        self.is_write_header_separator_row = True
        self.is_write_value_separator_row = True
        self.is_write_opening_row = True
        self.is_write_closing_row = True

        self.update_preprocessor(line_break_handling=LineBreakHandling.NOP)

        self._quoting_flags = copy.deepcopy(dp.NOT_QUOTING_FLAGS)

    def _create_styler(self, writer: AbstractTableWriter) -> StylerInterface:
        return AsciiDocStyler(writer)

    def _write_value_row(
        self, row: int, values: Sequence[str], value_dp_list: Sequence[DataProperty]
    ) -> None:
        self._write_row(
            row,
            [
                self.__modify_row_element(row, col_idx, value, value_dp)
                for col_idx, (value, value_dp), in enumerate(zip(values, value_dp_list))
            ],
        )

    def _get_opening_row_items(self) -> List[str]:
        cols = ", ".join(
            f"{get_align_char(col_dp.align)}{col_dp.ascii_char_width}"
            for col_dp in self._column_dp_list
        )
        rows = [f'[cols="{cols}", options="header"]']

        if typepy.is_not_null_string(self.table_name):
            rows.append("." + MultiByteStrDecoder(self.table_name).unicode_str)

        rows.append("|===")

        return ["\n".join(rows)]

    def _get_header_row_separator_items(self) -> List[str]:
        return [""]

    def _get_value_row_separator_items(self) -> List[str]:
        return self._get_header_row_separator_items()

    def _get_closing_row_items(self) -> List[str]:
        return ["|==="]

    def _get_header_format_string(self, col_dp: ColumnDataProperty, value_dp: DataProperty) -> str:
        return f"{get_align_char(Align.CENTER)}|{{}}"

    def __modify_row_element(self, row_idx: int, col_idx: int, value: str, value_dp: DataProperty):
        default_style = self._get_col_style(col_idx)
        col_dp = self._column_dp_list[col_idx]
        style = self._fetch_style_from_filter(row_idx, col_dp, value_dp, default_style)
        align = col_dp.align

        if style and style.align and style.align != align:
            forma_stirng = "{0:s}|{1:s}"
            align = style.align
        elif value_dp.align != align:
            forma_stirng = "{0:s}|{1:s}"
            align = value_dp.align
        else:
            forma_stirng = "|{1:s}"

        return forma_stirng.format(get_align_char(align), value)



import abc
from typing import Any, Optional, cast

from tcolorpy import tcolor

from ._font import FontSize, FontStyle, FontWeight
from ._style import Align, DecorationLine, Style, ThousandSeparator


_align_char_mapping = {
    Align.AUTO: "<",
    Align.LEFT: "<",
    Align.RIGHT: ">",
    Align.CENTER: "^",
}


def get_align_char(align: Align) -> str:
    return _align_char_mapping[align]


class StylerInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def apply(self, value: Any, style: Style) -> str:  # pragma: no cover
        raise NotImplementedError()

    @abc.abstractmethod
    def apply_align(self, value: str, style: Style) -> str:  # pragma: no cover
        raise NotImplementedError()

    @abc.abstractmethod
    def apply_terminal_style(self, value: str, style: Style) -> str:  # pragma: no cover
        raise NotImplementedError()

    @abc.abstractmethod
    def get_font_size(self, style: Style) -> Optional[str]:  # pragma: no cover
        raise NotImplementedError()

    @abc.abstractmethod
    def get_additional_char_width(self, style: Style) -> int:  # pragma: no cover
        raise NotImplementedError()


class AbstractStyler(StylerInterface):
    def __init__(self, writer):
        self._writer = writer
        self._font_size_map = self._get_font_size_map()

    def get_font_size(self, style: Style) -> Optional[str]:
        return self._font_size_map.get(style.font_size)

    def get_additional_char_width(self, style: Style) -> int:
        return 0

    def apply(self, value: Any, style: Style) -> str:
        return value

    def apply_align(self, value: str, style: Style) -> str:
        return value

    def apply_terminal_style(self, value: str, style: Style) -> str:
        return value

    def _get_font_size_map(self):
        return {}


class NullStyler(AbstractStyler):
    def get_font_size(self, style: Style) -> Optional[str]:
        return ""


class TextStyler(AbstractStyler):
    def apply_terminal_style(self, value: str, style: Style) -> str:
        if not self._writer.enable_ansi_escape:
            return value

        ansi_styles = []

        if style.decoration_line in (DecorationLine.STRIKE, DecorationLine.LINE_THROUGH):
            ansi_styles.append("strike")
        if style.decoration_line == DecorationLine.UNDERLINE:
            ansi_styles.append("underline")

        if style.font_weight == FontWeight.BOLD:
            ansi_styles.append("bold")

        if self._writer.colorize_terminal:
            return tcolor(value, color=style.color, bg_color=style.bg_color, styles=ansi_styles)

        return tcolor(value, styles=ansi_styles)

    def __get_align_format(self, style: Style) -> str:
        align_char = get_align_char(style.align)
        format_items = ["{:" + align_char]
        if style.padding is not None and style.padding > 0:
            format_items.append(str(style.padding))
        format_items.append("s}")

        return "".join(format_items)

    def apply_align(self, value: str, style: Style) -> str:
        return self.__get_align_format(style).format(value)

    def apply(self, value: Any, style: Style) -> str:
        if value:
            if style.thousand_separator == ThousandSeparator.SPACE:
                value = value.replace(",", " ")
            elif style.thousand_separator == ThousandSeparator.UNDERSCORE:
                value = value.replace(",", "_")

        return value


class HtmlStyler(TextStyler):
    def _get_font_size_map(self):
        return {
            FontSize.TINY: "font-size:x-small",
            FontSize.SMALL: "font-size:small",
            FontSize.MEDIUM: "font-size:medium",
            FontSize.LARGE: "font-size:large",
        }


class LatexStyler(TextStyler):
    class Command:
        BOLD = r"\bf"
        ITALIC = r"\it"

    def get_additional_char_width(self, style: Style) -> int:
        width = 0

        if self.get_font_size(style):
            width += len(cast(str, self.get_font_size(style)))

        if style.font_weight == FontWeight.BOLD:
            width += len(self.Command.BOLD)

        if style.font_style == FontStyle.ITALIC:
            width += len(self.Command.ITALIC)

        return width

    def apply(self, value: Any, style: Style) -> str:
        value = super().apply(value, style)
        if not value:
            return value

        font_size = self.get_font_size(style)
        item_list = []

        if font_size:
            item_list.append(font_size)

        if style.font_weight == FontWeight.BOLD:
            item_list.append(self.Command.BOLD)

        if style.font_style == FontStyle.ITALIC:
            item_list.append(self.Command.ITALIC)

        item_list.append(value)
        return " ".join(item_list)

    def _get_font_size_map(self):
        return {
            FontSize.TINY: r"\tiny",
            FontSize.SMALL: r"\small",
            FontSize.MEDIUM: r"\normalsize",
            FontSize.LARGE: r"\large",
        }


class MarkdownStyler(TextStyler):
    def get_additional_char_width(self, style: Style) -> int:
        width = 0

        if style.font_weight == FontWeight.BOLD:
            width += 4

        if style.font_style == FontStyle.ITALIC:
            width += 2

        return width

    def apply(self, value: Any, style: Style) -> str:
        value = super().apply(value, style)
        if not value:
            return value

        if style.font_weight == FontWeight.BOLD:
            value = f"**{value}**"

        if style.font_style == FontStyle.ITALIC:
            value = f"_{value}_"

        return value


class GFMarkdownStyler(MarkdownStyler):
    """
    A styler class for GitHub Flavored Markdown
    """

    def get_additional_char_width(self, style: Style) -> int:
        width = super().get_additional_char_width(style)

        if style.decoration_line in (DecorationLine.STRIKE, DecorationLine.LINE_THROUGH):
            width += 4

        return width

    def apply(self, value: Any, style: Style) -> str:
        value = super().apply(value, style)
        if not value:
            return value

        if style.decoration_line in (DecorationLine.STRIKE, DecorationLine.LINE_THROUGH):
            value = f"~~{value}~~"

        return value


class ReStructuredTextStyler(TextStyler):
    def get_additional_char_width(self, style: Style) -> int:
        from ..writer import RstCsvTableWriter

        width = 0

        if style.font_weight == FontWeight.BOLD:
            width += 4
        elif style.font_style == FontStyle.ITALIC:
            width += 2

        if (
            style.thousand_separator == ThousandSeparator.COMMA
            and self._writer.format_name == RstCsvTableWriter.FORMAT_NAME
        ):
            width += 2

        return width

    def apply(self, value: Any, style: Style) -> str:
        from ..writer import RstCsvTableWriter

        value = super().apply(value, style)
        if not value:
            return value

        if style.font_weight == FontWeight.BOLD:
            value = f"**{value}**"
        elif style.font_style == FontStyle.ITALIC:
            # in reStructuredText, some custom style definition will be required to
            # set for both bold and italic (currently not supported)
            value = f"*{value}*"

        if (
            style.thousand_separator == ThousandSeparator.COMMA
            and self._writer.format_name == RstCsvTableWriter.FORMAT_NAME
        ):
            value = f'"{value}"'

        return value



import copy
import re
from typing import List, Sequence

import dataproperty as dp
import typepy
from dataproperty import ColumnDataProperty, DataProperty, LineBreakHandling
from mbstrdecoder import MultiByteStrDecoder

from ...style import Align
from ...style._styler import get_align_char
from ._text_writer import TextTableWriter


class MediaWikiTableWriter(TextTableWriter):
    """
    A table writer class for `MediaWiki <https://www.mediawiki.org/wiki/MediaWiki>`__ format.

        :Example:
            :ref:`example-mediawiki-table-writer`
    """

    FORMAT_NAME = "mediawiki"
    __RE_TABLE_SEQUENCE = re.compile(r"^[\s]+[*|#]+")

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    @property
    def support_split_write(self) -> bool:
        return True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.column_delimiter = "\n"

        self.is_padding = False
        self.is_write_header_separator_row = True
        self.is_write_value_separator_row = True
        self.is_write_opening_row = True
        self.is_write_closing_row = True

        self.update_preprocessor(line_break_handling=LineBreakHandling.NOP)

        self._quoting_flags = copy.deepcopy(dp.NOT_QUOTING_FLAGS)

        # experimental: This attribute may change in the future release
        self.table_style = kwargs.get("table_style", "")

    def _write_header(self) -> None:
        if not self.is_write_header:
            return

        if typepy.is_not_null_string(self.table_name):
            self._write_line("|+" + MultiByteStrDecoder(self.table_name).unicode_str)

        super()._write_header()

    def _write_value_row(
        self, row: int, values: Sequence[str], value_dp_list: Sequence[DataProperty]
    ) -> None:
        self._write_row(
            row,
            [
                self.__modify_table_element(value, value_dp)
                for value, value_dp, in zip(values, value_dp_list)
            ],
        )

    def _get_opening_row_items(self) -> List[str]:
        item = '{| class="wikitable"'
        if self.table_style:
            item += f' style="{self.table_style}"'

        return [item]

    def _get_header_row_separator_items(self) -> List[str]:
        return ["|-"]

    def _get_value_row_separator_items(self) -> List[str]:
        return self._get_header_row_separator_items()

    def _get_closing_row_items(self) -> List[str]:
        return ["|}"]

    def _get_header_format_string(self, col_dp: ColumnDataProperty, value_dp: DataProperty) -> str:
        return "! {{:{:s}{:s}}}".format(
            get_align_char(Align.CENTER), str(self._get_padding_len(col_dp, value_dp))
        )

    def __modify_table_element(self, value: str, value_dp: DataProperty):
        if value_dp.align is Align.LEFT:
            forma_stirng = "| {1:s}"
        else:
            forma_stirng = '| style="text-align:{0:s}"| {1:s}'

        if self.__RE_TABLE_SEQUENCE.search(value) is not None:
            value = "\n" + value.lstrip()

        return forma_stirng.format(value_dp.align.align_string, value)



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

"""
Module elastic_blast.jobs

Generate Kubernetes job YAML files from template

Author: Victor Joukov joukovv@ncbi.nlm.nih.gov
"""

import os
import re
from typing import List
from pkg_resources import resource_string

from .filehelper import open_for_read, open_for_write
from .subst import substitute_params
from .constants import ELB_DFLT_BLAST_JOB_TEMPLATE, ELB_LOCAL_SSD_BLAST_JOB_TEMPLATE
from .elb_config import ElasticBlastConfig

def read_job_template(template_name=ELB_DFLT_BLAST_JOB_TEMPLATE, cfg: ElasticBlastConfig = None):
    """ Read job template file or resource
    Parameters:
        template_name - name of file to read or default resource
    Returns:
        string with job template text
    """
    if cfg and cfg.cluster.use_local_ssd:
        template_name = ELB_LOCAL_SSD_BLAST_JOB_TEMPLATE
    resource_prefix = 'resource:'
    resource_prefix_len = len(resource_prefix)
    if template_name[:resource_prefix_len] == resource_prefix:
        template_name = template_name[resource_prefix_len:]
        return resource_string('elastic_blast', template_name).decode()
    with open_for_read(template_name) as f:
        return f.read()


re_batch_num = re.compile(r'[^0-9]+([0-9]{3,})')


def _write_job_file(job_path, job_prefix, job_template, query_fqn, njob, **subs):
    """ Write YAML job file from template making substitutions
        internal function
    Parameters:
        job_path: path to which write job files
        job_prefix: name prefix for job file
        job_template: string with contents of job file with variables to substitute
        query_fqn: fully qualified name of query file
        njob: ordinal number of a job
        subs: other substitution variables
    Result:
        Job file name
    """
    if not job_template:
        return None
    job_file_name = os.path.join(job_path, f'{job_prefix}{njob:03d}.yaml')
    query_path = os.path.dirname(query_fqn)
    query = os.path.splitext(os.path.basename(query_fqn))[0]
    # Try to recover batch number from file name, if not available use njob
    mo = re_batch_num.match(query)
    if mo:
        query_num = mo.group(1)
    else:
        query_num = f'{njob:03d}'

    map_obj = {}
    for k, v in subs.items():
        map_obj[k] = v
    map_obj['QUERY'] = query
    map_obj['QUERY_FQN'] = query_fqn
    map_obj['QUERY_PATH'] = query_path
    map_obj['QUERY_NUM'] = query_num
    map_obj['JOB_NUM'] = query_num

    s = substitute_params(job_template, map_obj)
    with open_for_write(job_file_name) as f:
        f.write(s)
    return job_file_name


def write_job_files(job_path: str, job_prefix: str, job_template: str, queries: List[str], **subs):
    """ Write YAML job files from template making substitutions
    Parameters:
        job_path: path to which write job files
        job_prefix: name prefix for job file
        job_template: string with contents of job file with variables to substitute
        queries: list of query file names to substitute in job files
        subs: other substitution variables
    Result:
        List of job file names
    """
    if not job_template:
        return []
    jobs = []
    for njob, query in enumerate(queries):
        subs['BLAST_ELB_BATCH_NUM'] = str(njob)
        job = _write_job_file(job_path, job_prefix,
                              job_template, query, njob, **subs)
        jobs.append(job)
    return jobs



import copy
from textwrap import indent
from typing import List

import dataproperty
import typepy
from dataproperty import ColumnDataProperty, DataProperty
from mbstrdecoder import MultiByteStrDecoder
from tabledata import to_value_matrix
from typepy import Typecode

from .._msgfy import to_error_message
from ._common import serialize_dp
from ._text_writer import IndentationTextTableWriter


try:
    import simplejson as json
except ImportError:
    import json  # type: ignore


class JsonTableWriter(IndentationTextTableWriter):
    """
    A table writer class for JSON format.

        :Examples:
            :ref:`example-json-table-writer`

    .. py:method:: write_table

        |write_table| with JSON format.

        Args:
            indent (Optional[int]):
                Indent level of an output.
                Interpretation of indent level value differ format to format.
                Some writer classes may ignore this value.
                Defaults to 4.

            sort_keys (Optional[bool]):
                If |True|, the output of dictionaries will be sorted by key.
                Defaults to |False|.

        Examples:
            :ref:`example-json-table-writer`

        .. note::
            Specific values in the tabular data are converted when writing:

            - |None|: written as ``null``
            - |inf|: written as ``Infinity``
            - |nan|: written as ``NaN``
    """

    FORMAT_NAME = "json"

    @property
    def format_name(self) -> str:
        return self.FORMAT_NAME

    @property
    def support_split_write(self) -> bool:
        return True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.set_indent_level(4)

        self.is_formatting_float = False
        self.is_write_opening_row = True
        self.is_write_closing_row = True
        self.char_right_side_row = ","
        self.char_opening_row_cross_point = ""
        self.char_closing_row_cross_point = ""

        self._is_require_header = True
        self._dp_extractor.type_value_map = {
            Typecode.INFINITY: "Infinity",
            Typecode.NAN: "NaN",
        }
        self._dp_extractor.update_strict_level_map({Typecode.BOOL: typepy.StrictLevel.MAX})
        self._quoting_flags = copy.deepcopy(dataproperty.NOT_QUOTING_FLAGS)

        self._init_cross_point_maps()

    def write_null_line(self) -> None:
        self._verify_stream()
        self.stream.write("\n")

    def _write_table(self, **kwargs) -> None:
        sort_keys = kwargs.get("sort_keys", False)

        self._preprocess()

        with self._logger:
            self._write_opening_row()

            json_text_list = []
            for json_data in self._table_value_matrix:
                json_text = json.dumps(
                    json_data, indent=self._indent_level, ensure_ascii=False, sort_keys=sort_keys
                )
                json_text_list.append(json_text)

            joint_text = self.char_right_side_row + "\n"
            json_text = joint_text.join(json_text_list)
            if all([not self.is_write_closing_row, typepy.is_not_null_string(json_text)]):
                json_text += joint_text

            self.stream.write(indent(json_text, " " * self._indent_level))
            self._write_closing_row()

    def _to_row_item(self, row_idx: int, col_dp: ColumnDataProperty, value_dp: DataProperty) -> str:
        return value_dp.to_str()

    def _preprocess_table_dp(self) -> None:
        if self._is_complete_table_dp_preprocess:
            return

        self._logger.logger.debug("_preprocess_table_dp")

        try:
            self._table_value_dp_matrix = self._dp_extractor.to_dp_matrix(
                to_value_matrix(self.headers, self.value_matrix)
            )
        except TypeError as e:
            self._logger.logger.debug(to_error_message(e))
            self._table_value_dp_matrix = []

        self._is_complete_table_dp_preprocess = True

    def _preprocess_header(self) -> None:
        if self._is_complete_header_preprocess:
            return

        self._logger.logger.debug("_preprocess_header")

        self._table_headers = [
            header_dp.to_str() for header_dp in self._dp_extractor.to_header_dp_list()
        ]

        self._is_complete_header_preprocess = True

    def _preprocess_value_matrix(self) -> None:
        if self._is_complete_value_matrix_preprocess:
            return

        self._table_value_matrix = [
            dict(zip(self._table_headers, [serialize_dp(dp) for dp in dp_list]))
            for dp_list in self._table_value_dp_matrix
        ]

        self._is_complete_value_matrix_preprocess = True

    def _get_opening_row_items(self) -> List[str]:
        if typepy.is_not_null_string(self.table_name):
            return [f'{{ "{MultiByteStrDecoder(self.table_name).unicode_str:s}" : [']

        return ["["]

    def _get_closing_row_items(self) -> List[str]:
        if typepy.is_not_null_string(self.table_name):
            return ["\n]}"]

        return ["\n]"]



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import abc
import os
from typing import Optional, Tuple, cast

from ._common import PathType, Platform, PlatformType, normalize_platform, unprintable_ascii_chars
from .error import ReservedNameError, ValidationError
from .handler import Handler, return_null_string


DEFAULT_MIN_LEN = 1


class BaseFile:
    _INVALID_PATH_CHARS = "".join(unprintable_ascii_chars)
    _INVALID_FILENAME_CHARS = _INVALID_PATH_CHARS + "/"
    _INVALID_WIN_PATH_CHARS = _INVALID_PATH_CHARS + ':*?"<>|\t\n\r\x0b\x0c'
    _INVALID_WIN_FILENAME_CHARS = _INVALID_FILENAME_CHARS + _INVALID_WIN_PATH_CHARS + "\\"

    _ERROR_MSG_TEMPLATE = "invalid char found: invalids=({invalid}), value={value}"

    @property
    def platform(self) -> Platform:
        return self.__platform

    @property
    def reserved_keywords(self) -> Tuple[str, ...]:
        return tuple()

    @property
    def min_len(self) -> int:
        return self._min_len

    @property
    def max_len(self) -> int:
        return self._max_len

    def __init__(
        self,
        min_len: int,
        max_len: int,
        check_reserved: bool,
        null_value_handler: Optional[Handler] = None,
        platform_max_len: Optional[int] = None,
        platform: PlatformType = None,
    ) -> None:
        self.__platform = normalize_platform(platform)
        self._check_reserved = check_reserved

        if null_value_handler is None:
            null_value_handler = return_null_string
        self._null_value_handler = null_value_handler

        if min_len <= 0:
            min_len = DEFAULT_MIN_LEN
        self._min_len = max(min_len, 1)

        if platform_max_len is None:
            platform_max_len = self._get_default_max_path_len()

        if max_len <= 0:
            self._max_len = platform_max_len
        else:
            self._max_len = cast(int, max_len)

        self._max_len = min(self._max_len, platform_max_len)
        self._validate_max_len()

    def _is_posix(self) -> bool:
        return self.platform == Platform.POSIX

    def _is_universal(self) -> bool:
        return self.platform == Platform.UNIVERSAL

    def _is_linux(self) -> bool:
        return self.platform == Platform.LINUX

    def _is_windows(self) -> bool:
        return self.platform == Platform.WINDOWS

    def _is_macos(self) -> bool:
        return self.platform == Platform.MACOS

    def _validate_max_len(self) -> None:
        if self.max_len < 1:
            raise ValueError("max_len must be greater or equals to one")

        if self.min_len > self.max_len:
            raise ValueError("min_len must be lower than max_len")

    def _get_default_max_path_len(self) -> int:
        if self._is_linux():
            return 4096

        if self._is_windows():
            return 260

        if self._is_posix() or self._is_macos():
            return 1024

        return 260  # universal


class AbstractValidator(BaseFile, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def validate(self, value: PathType) -> None:  # pragma: no cover
        pass

    def is_valid(self, value: PathType) -> bool:
        try:
            self.validate(value)
        except (TypeError, ValidationError):
            return False

        return True

    def _is_reserved_keyword(self, value: str) -> bool:
        return value in self.reserved_keywords


class AbstractSanitizer(BaseFile, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def sanitize(self, value: PathType, replacement_text: str = "") -> PathType:  # pragma: no cover
        pass


class BaseValidator(AbstractValidator):
    def _validate_reserved_keywords(self, name: str) -> None:
        if not self._check_reserved:
            return

        root_name = self.__extract_root_name(name)
        if self._is_reserved_keyword(root_name.upper()):
            raise ReservedNameError(
                f"'{root_name}' is a reserved name",
                reusable_name=False,
                reserved_name=root_name,
                platform=self.platform,
            )

    @staticmethod
    def __extract_root_name(path: str) -> str:
        return os.path.splitext(os.path.basename(path))[0]



from decimal import Decimal
from typing import Any, Callable, Mapping, Optional, Type, Union

from typepy import (
    Bool,
    DateTime,
    Dictionary,
    Infinity,
    Integer,
    IpAddress,
    List,
    Nan,
    NoneType,
    NullString,
    RealNumber,
    String,
    Typecode,
)
from typepy.type import AbstractType


StrictLevelMap = Mapping[Union[str, Typecode], int]
TypeHint = Optional[Type[AbstractType]]
TransFunc = Callable[[Any], Any]

FloatType = Union[Type[Decimal], Type[float]]

_type_hint_map = {
    # high frequently used types
    "int": Integer,
    "float": RealNumber,
    "realnumber": RealNumber,
    "str": String,
    # low frequently used types
    "bool": Bool,
    "datetime": DateTime,
    "dict": Dictionary,
    "inf": Infinity,
    "ip": IpAddress,
    "list": List,
    "nan": Nan,
    "none": NoneType,
    "nullstr": NullString,
}


def normalize_type_hint(type_hint: Union[str, TypeHint]) -> TypeHint:
    if not type_hint:
        return None

    if not isinstance(type_hint, str):
        return type_hint

    type_hint = type_hint.strip().casefold()
    for key, value in _type_hint_map.items():
        if type_hint.startswith(key):
            return value

    raise ValueError(f"unknown typehint: {type_hint}")



# -*- coding: utf-8 -*-
# Copyright (c) 2014 Rackspace
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with functions to normalize components."""
import re

from . import compat
from . import misc


def normalize_scheme(scheme):
    """Normalize the scheme component."""
    return scheme.lower()


def normalize_authority(authority):
    """Normalize an authority tuple to a string."""
    userinfo, host, port = authority
    result = ""
    if userinfo:
        result += normalize_percent_characters(userinfo) + "@"
    if host:
        result += normalize_host(host)
    if port:
        result += ":" + port
    return result


def normalize_username(username):
    """Normalize a username to make it safe to include in userinfo."""
    return compat.urlquote(username)


def normalize_password(password):
    """Normalize a password to make safe for userinfo."""
    return compat.urlquote(password)


def normalize_host(host):
    """Normalize a host string."""
    if misc.IPv6_MATCHER.match(host):
        percent = host.find("%")
        if percent != -1:
            percent_25 = host.find("%25")

            # Replace RFC 4007 IPv6 Zone ID delimiter '%' with '%25'
            # from RFC 6874. If the host is '[<IPv6 addr>%25]' then we
            # assume RFC 4007 and normalize to '[<IPV6 addr>%2525]'
            if (
                percent_25 == -1
                or percent < percent_25
                or (percent == percent_25 and percent_25 == len(host) - 4)
            ):
                host = host.replace("%", "%25", 1)

            # Don't normalize the casing of the Zone ID
            return host[:percent].lower() + host[percent:]

    return host.lower()


def normalize_path(path):
    """Normalize the path string."""
    if not path:
        return path

    path = normalize_percent_characters(path)
    return remove_dot_segments(path)


def normalize_query(query):
    """Normalize the query string."""
    if not query:
        return query
    return normalize_percent_characters(query)


def normalize_fragment(fragment):
    """Normalize the fragment string."""
    if not fragment:
        return fragment
    return normalize_percent_characters(fragment)


PERCENT_MATCHER = re.compile("%[A-Fa-f0-9]{2}")


def normalize_percent_characters(s):
    """All percent characters should be upper-cased.

    For example, ``"%3afoo%DF%ab"`` should be turned into ``"%3Afoo%DF%AB"``.
    """
    matches = set(PERCENT_MATCHER.findall(s))
    for m in matches:
        if not m.isupper():
            s = s.replace(m, m.upper())
    return s


def remove_dot_segments(s):
    """Remove dot segments from the string.

    See also Section 5.2.4 of :rfc:`3986`.
    """
    # See http://tools.ietf.org/html/rfc3986#section-5.2.4 for pseudo-code
    segments = s.split("/")  # Turn the path into a list of segments
    output = []  # Initialize the variable to use to store output

    for segment in segments:
        # '.' is the current directory, so ignore it, it is superfluous
        if segment == ".":
            continue
        # Anything other than '..', should be appended to the output
        elif segment != "..":
            output.append(segment)
        # In this case segment == '..', if we can, we should pop the last
        # element
        elif output:
            output.pop()

    # If the path starts with '/' and the output is empty or the first string
    # is non-empty
    if s.startswith("/") and (not output or output[0]):
        output.insert(0, "")

    # If the path starts with '/.' or '/..' ensure we add one more empty
    # string to add a trailing '/'
    if s.endswith(("/.", "/..")):
        output.append("")

    return "/".join(output)


def encode_component(uri_component, encoding):
    """Encode the specific component in the provided encoding."""
    if uri_component is None:
        return uri_component

    # Try to see if the component we're encoding is already percent-encoded
    # so we can skip all '%' characters but still encode all others.
    percent_encodings = len(
        PERCENT_MATCHER.findall(compat.to_str(uri_component, encoding))
    )

    uri_bytes = compat.to_bytes(uri_component, encoding)
    is_percent_encoded = percent_encodings == uri_bytes.count(b"%")

    encoded_uri = bytearray()

    for i in range(0, len(uri_bytes)):
        # Will return a single character bytestring on both Python 2 & 3
        byte = uri_bytes[i : i + 1]
        byte_ord = ord(byte)
        if (is_percent_encoded and byte == b"%") or (
            byte_ord < 128 and byte.decode() in misc.NON_PCT_ENCODED
        ):
            encoded_uri.extend(byte)
            continue
        encoded_uri.extend("%{0:02x}".format(byte_ord).encode().upper())

    return encoded_uri.decode(encoding)



from .._connection import *
from .._events import *
from .._state import *


def get_all_events(conn):
    got_events = []
    while True:
        event = conn.next_event()
        if event in (NEED_DATA, PAUSED):
            break
        got_events.append(event)
        if type(event) is ConnectionClosed:
            break
    return got_events


def receive_and_get(conn, data):
    conn.receive_data(data)
    return get_all_events(conn)


# Merges adjacent Data events, converts payloads to bytestrings, and removes
# chunk boundaries.
def normalize_data_events(in_events):
    out_events = []
    for event in in_events:
        if type(event) is Data:
            event.data = bytes(event.data)
            event.chunk_start = False
            event.chunk_end = False
        if out_events and type(out_events[-1]) is type(event) is Data:
            out_events[-1].data += event.data
        else:
            out_events.append(event)
    return out_events


# Given that we want to write tests that push some events through a Connection
# and check that its state updates appropriately... we might as make a habit
# of pushing them through two Connections with a fake network link in
# between.
class ConnectionPair:
    def __init__(self):
        self.conn = {CLIENT: Connection(CLIENT), SERVER: Connection(SERVER)}
        self.other = {CLIENT: SERVER, SERVER: CLIENT}

    @property
    def conns(self):
        return self.conn.values()

    # expect="match" if expect=send_events; expect=[...] to say what expected
    def send(self, role, send_events, expect="match"):
        if not isinstance(send_events, list):
            send_events = [send_events]
        data = b""
        closed = False
        for send_event in send_events:
            new_data = self.conn[role].send(send_event)
            if new_data is None:
                closed = True
            else:
                data += new_data
        # send uses b"" to mean b"", and None to mean closed
        # receive uses b"" to mean closed, and None to mean "try again"
        # so we have to translate between the two conventions
        if data:
            self.conn[self.other[role]].receive_data(data)
        if closed:
            self.conn[self.other[role]].receive_data(b"")
        got_events = get_all_events(self.conn[self.other[role]])
        if expect == "match":
            expect = send_events
        if not isinstance(expect, list):
            expect = [expect]
        assert got_events == expect
        return data



from decimal import Decimal
from typing import Any

from dataproperty import DataProperty
from typepy import Typecode


def bool_to_str(value) -> str:
    if value is True:
        return "true"
    if value is False:
        return "false"

    return value


def serialize_dp(dp: DataProperty) -> Any:
    if dp.typecode in (Typecode.REAL_NUMBER, Typecode.INFINITY, Typecode.NAN) and isinstance(
        dp.data, Decimal
    ):
        return float(dp.data)

    if dp.typecode == Typecode.DATETIME:
        return dp.to_str()

    return dp.data



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import re
from decimal import Decimal

from ._checker import CheckerFactory, TypeCheckerBase, TypeCheckerDelegator
from ._common import isinf, isnan, isstring


RE_E = re.compile("[eE]")
RE_SCIENTIFIC_NOTATION = re.compile(r"^-?\d+(?:\.\d*)?[eE][+\-]?\d{,2}$")


class RealNumberTypeCheckerStrictLevel0(TypeCheckerBase):
    def is_instance(self):
        return isinstance(self._value, (float, Decimal))

    def is_exclude_instance(self):
        return isinstance(self._value, bool) or isnan(self._value) or isinf(self._value)

    def is_valid_after_convert(self, converted_value):
        return not isinf(converted_value) and not isnan(converted_value)


class RealNumberTypeCheckerStrictLevel1(RealNumberTypeCheckerStrictLevel0):
    def is_instance(self):
        return super().is_instance() and not float(self._value).is_integer()

    def is_exclude_instance(self):
        if (
            isinstance(self._value, str)
            and RE_E.search(self._value)
            and RE_SCIENTIFIC_NOTATION.search(self._value) is None
        ):
            return True

        return isinstance(self._value, int) or super().is_exclude_instance()

    def is_valid_after_convert(self, converted_value):
        return not float(converted_value).is_integer()


class RealNumberTypeCheckerStrictLevel2(RealNumberTypeCheckerStrictLevel1):
    def is_exclude_instance(self):
        return super().is_exclude_instance() or isstring(self._value)


_factory = CheckerFactory(
    checker_mapping={
        0: RealNumberTypeCheckerStrictLevel0,
        1: RealNumberTypeCheckerStrictLevel1,
        2: RealNumberTypeCheckerStrictLevel2,
    }
)


class RealNumberTypeChecker(TypeCheckerDelegator):
    def __init__(self, value, strict_level):
        super().__init__(value=value, checker_factory=_factory, strict_level=strict_level)



#!/usr/bin/env python3
#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

"""
Split FASTA file into smaller chunks

File can be on local filesystem, available by URL,
or reside in GCP bucket (GS). File can be compressed and/or
archived (contents of all files in the archive is treated as
one large merged file). Following combinations are recognized:
.gz, .tar, .tar.gz, .tgz, .tar.bz2 .

Also creates YAML files for each generated piece from template,
substituting several variable.

For details see README file

Author: Victor Joukov joukovv@ncbi.nlm.nih.gov
"""

import sys
import argparse
from tarfile import ReadError
from elastic_blast.filehelper import open_for_read, open_for_write, copy_to_bucket
from elastic_blast.split import FASTAReader
from elastic_blast.jobs import write_job_files
from elastic_blast.constants import ELB_QUERY_BATCH_FILE_PREFIX


DEFAULT_BATCH_LEN    = 5000000
DEFAULT_OUT_PATH     = 'batches'
DEFAULT_RES_PATH     = 'results'
DEFAULT_JOB_PATH     = 'jobs'

manifest_file = sys.stdout


def parse_arguments():
    parser = argparse.ArgumentParser(description="Split FASTA file")
    parser.add_argument('input', help='input FASTA file, possible gzipped')
    parser.add_argument('-l', '--batch_len', type=int, default=DEFAULT_BATCH_LEN,
        help='batch length')
    parser.add_argument('-o', '--output',    default=DEFAULT_OUT_PATH,
        help='output path for batch FASTA files')
    parser.add_argument('-r', '--results',   default=DEFAULT_RES_PATH,
        help='output path for BLAST results')
    parser.add_argument('-j', '--job_path',  default=DEFAULT_JOB_PATH,
        help='output path for job YAML files')
    parser.add_argument('-t', '--template',  default='',
        help='YAML template')
    parser.add_argument('-s', '--subs',  default='',
        help='Variable substitutes in form var1=vale1,var2=value2 ...')
    parser.add_argument('-m', '--manifest',  default='',
        help='manifest file to write')
    parser.add_argument('-c', '--count',  default='',
        help='file to report total number of bases/residues in input file')
    parser.add_argument("-n", "--dry-run", action='store_true', 
                        help="Do not run any commands, just show what would be executed")
    return parser.parse_args()

def main():
    global manifest_file
    args = parse_arguments()
    input_path   = args.input
    out_path     = args.output
    res_path     = args.results
    job_path     = args.job_path
    batch_len    = args.batch_len
    job_template = args.template
    manifest     = args.manifest
    count_file   = args.count
    dry_run      = args.dry_run
    job_template_text = ''
    try:
        if job_template:
            with open_for_read(job_template) as f:
                job_template_text = f.read()
    except FileNotFoundError as e:
        print(e, "for --template parameter", file=sys.stderr)
        return 1
    # Convert string of form key1=val1,key2=val2 into dictionary { 'key1' : 'val1', 'key2' : 'val2' }
    subs = { key: value for key, value in filter(lambda x: len(x) == 2 and x[0], map(lambda x: x.split('='), args.subs.split(','))) }
    subs['RESULTS'] = res_path
    total_count = 0
    try:
        with open_for_read(input_path) as s:
            reader = FASTAReader(s, batch_len, out_path)
            total_count, queries = reader.read_and_cut()
        jobs = write_job_files(job_path, ELB_QUERY_BATCH_FILE_PREFIX, job_template_text, queries, **subs)
        if count_file:
            if count_file == '-':
                sys.stdout.write(str(total_count)+'\n')
            else:
                with open_for_write(count_file) as f:
                    f.write(str(total_count))
        if jobs and manifest:
            manifest_text = '\n'.join(jobs)+'\n'
            if manifest == '-':
                sys.stdout.write(manifest_text)
            else:
                with open_for_write(manifest) as manifest_file:
                    manifest_file.write(manifest_text)
    except FileNotFoundError as e:
        print(e, "for input file", file=sys.stderr)
        return 2
    except PermissionError as e:
        print(e, file=sys.stderr)
        return 3
    except UnicodeDecodeError as e:
        print(e, "\nPossibly missing .gz or .tar extension for compressed or archived file", file=sys.stderr)
        return 4
    except OSError as e:
        # If .gz extension is present on not gzipped file
        print(e, file=sys.stderr)
        return 5
    except ReadError as e:
        print(e, "\nProbably not a tar file", file=sys.stderr)
        return 6
    except Exception as e:
        # If the file is empty
        print(e, file=sys.stderr)
        return 7
    copy_to_bucket(dry_run)
    return 0


if __name__ == "__main__":
    sys.exit(main())
